"""
This file contains some basic function for weno.
"""


def weno5_LR(vmm, vm, vo, vp, vpp):
    """
    weno5_LR 5 order weno method

    Args:
        vmm ([type]): left left cell
        vm ([type]): left cell
        vo ([type]): middle cell
        vp ([type]): right cell
        vpp ([type]): right right cell

    Returns:
        Tuple[fl,fr]: fl and fr are reconstructions at left and right boundary of middle cell.
    """
    # reconstruction as x_{j+1/2}
    # FLUX SIZE: [...,W]
    # vmm = fluxP[..., :-4]
    # vm = fluxP[..., 1:-3]
    # vo = fluxP[..., 2:-2]
    # vp = fluxP[..., 3:-1]
    # vpp = fluxP[..., 4:]

    # Smooth Indicators (Beta factors)
    B0 = 13 / 12 * (vmm - 2 * vm + vo) ** 2 + 1 / 4 * (vmm - 4 * vm + 3 * vo) ** 2
    B1 = 13 / 12 * (vm - 2 * vo + vp) ** 2 + 1 / 4 * (vm - vp) ** 2
    B2 = 13 / 12 * (vo - 2 * vp + vpp) ** 2 + 1 / 4 * (3 * vo - 4 * vp + vpp) ** 2

    # Constants
    d0p = 3 / 10
    d1p = 6 / 10
    d2p = 1 / 10
    epsilon = 1e-6

    # Alpha weights
    alpha0p = d0p / (epsilon + B0) ** 2
    alpha1p = d1p / (epsilon + B1) ** 2
    alpha2p = d2p / (epsilon + B2) ** 2
    alphasump = alpha0p + alpha1p + alpha2p

    # ENO stencils weigths
    w0p = alpha0p / alphasump
    w1p = alpha1p / alphasump
    w2p = alpha2p / alphasump

    # Numerical Flux at cell boundary, $u_{i+1/2}^{+}$;
    fluxL = (
        w0p * (-vmm + 5 * vm + 2 * vo) / 6
        + w1p * (2 * vm + 5 * vo - vp) / 6
        + w2p * (11 * vo - 7 * vp + 2 * vpp) / 6
    )

    d0n = 1 / 10
    d1n = 6 / 10
    d2n = 3 / 10
    epsilon = 1e-6

    # Alpha weights
    alpha0n = d0n / (epsilon + B0) ** 2
    alpha1n = d1n / (epsilon + B1) ** 2
    alpha2n = d2n / (epsilon + B2) ** 2
    alphasumn = alpha0n + alpha1n + alpha2n

    # ENO stencils weigths
    w0n = alpha0n / alphasumn
    w1n = alpha1n / alphasumn
    w2n = alpha2n / alphasumn

    # Numerical Flux at cell boundary, $u_{i+1/2}^{-}$;
    fluxR = (
        w0n * (2 * vmm - 7 * vm + 11 * vo) / 6
        + w1n * (-vm + 5 * vo + 2 * vp) / 6
        + w2n * (2 * vo + 5 * vp - vpp) / 6
    )

    return fluxL, fluxR

def rescon_LR(vmm,vm,vo,vp,vpp):
    # reconstruction as x_{j+1/2}
    # FLUX SIZE: [...,W]
    # vmm = fluxP[..., :-4]
    # vm = fluxP[..., 1:-3]
    # vo = fluxP[..., 2:-2]
    # vp = fluxP[..., 3:-1]
    # vpp = fluxP[..., 4:]

    # Smooth Indicators (Beta factors)
    B0 = 13 / 12 * (vmm - 2 * vm + vo) ** 2 + 1 / 4 * (vmm - 4 * vm + 3 * vo) ** 2
    B1 = 13 / 12 * (vm - 2 * vo + vp) ** 2 + 1 / 4 * (vm - vp) ** 2
    B2 = 13 / 12 * (vo - 2 * vp + vpp) ** 2 + 1 / 4 * (3 * vo - 4 * vp + vpp) ** 2

    # Constants
    d0p = 3 / 10
    d1p = 6 / 10
    d2p = 1 / 10
    epsilon = 1e-6

    # Alpha weights
    alpha0p = d0p / (epsilon + B0) ** 2
    alpha1p = d1p / (epsilon + B1) ** 2
    alpha2p = d2p / (epsilon + B2) ** 2
    alphasump = alpha0p + alpha1p + alpha2p

    # ENO stencils weigths
    w0p = alpha0p / alphasump
    w1p = alpha1p / alphasump
    w2p = alpha2p / alphasump

    # w0p= d0p
    # w1p= d1p
    # w2p= d2p

    # w0p= 0
    # w1p= 1
    # w2p= 0

    # Numerical Flux at cell boundary, $u_{i+1/2}^{+}$;
    fluxL = (
        w0p * (-vmm + 5 * vm + 2 * vo) / 6
        + w1p * (2 * vm + 5 * vo - vp) / 6
        + w2p * (11 * vo - 7 * vp + 2 * vpp) / 6
    )

    d0n = 1 / 10
    d1n = 6 / 10
    d2n = 3 / 10
    epsilon = 1e-6

    # Alpha weights
    alpha0n = d0n / (epsilon + B0) ** 2
    alpha1n = d1n / (epsilon + B1) ** 2
    alpha2n = d2n / (epsilon + B2) ** 2
    alphasumn = alpha0n + alpha1n + alpha2n

    # ENO stencils weigths
    w0n = alpha0n / alphasumn
    w1n = alpha1n / alphasumn
    w2n = alpha2n / alphasumn

    # w0n= d0n
    # w1n= d1n
    # w2n= d2n

    # w0n= 0
    # w1n= 1
    # w2n= 0

    # Numerical Flux at cell boundary, $u_{i+1/2}^{-}$;
    fluxR = (
        w0n * (2 * vmm - 7 * vm + 11 * vo) / 6
        + w1n * (-vm + 5 * vo + 2 * vp) / 6
        + w2n * (2 * vo + 5 * vp - vpp) / 6
    )
   
    return fluxL, fluxR


import numpy as np

# u_t = f(u) + g(u)
# f(u) non-stiff, explicit
# g(u) stiff, implicit

# Butcher tableau
# c|A
# ---
#  |b

# g(u,args)  BGK part: g=1/tau(M(f)-f)
# f(u,args)  flux and other collision part

# Implicit-Explicit schemes for BGK kinetic equations

# y size [B,C,W]

class IMEX_RK:
    def __init__(self, b1, A1, b2, A2) -> None:
        """
        __init__ [summary]

         u_t = f(u) + g(u)
        # f(u) non-stiff, explicit
        # g(u) stiff, implicit

        Args:
            b1 ([type]): [description]
            A1 ([type]): [description]
            b2 ([type]): [description]
            A2 ([type]): [description]

        Raises:
            ValueError: A1,b1,A2,b2 shape must be consistent.
        """
        self.Stage = len(b1)
        if len(b2) != self.Stage:
            raise ValueError("Invalid Butcher tableau")
        if A1.shape[0] != A1.shape[1] or A1.shape[0] != self.Stage:
            raise ValueError("Invalid Butcher tableau")
        if A2.shape[0] != A2.shape[1] or A2.shape[0] != self.Stage:
            raise ValueError("Invalid Butcher tableau")
        self.b1 = b1
        self.A1 = A1
        self.b2 = b2
        self.A2 = A2
        # Bucher(b1,A1) is for the f
        # Bucher(b2,A2) is for the g

    def __call__(self, f, y0, h, getTau, getU, Maxwellian):
        y_list = []
        f_list = []
        g_list = []
        U_list = []
        tau_list = []
        for s in range(self.Stage):
            #print(y_list,f_list,self.A1[s, :])
            #print([self.A1[s, i] * f_list[i] for i in range(s)])
            res_f = sum([self.A1[s, i] * f_list[i] for i in range(s)])
            res_g = sum([self.A2[s, i] * g_list[i] for i in range(s)])
            B = y0 + h * res_f + h * res_g
            U = getU(B)
            U_list.append(U)
            tau = getTau(U)
            tau_list.append(tau)
            # print(tau.shape,B.shape,Maxwellian(U).shape)
            y = (tau * B + h * self.A2[s, s] * Maxwellian(U)) / (
                tau + h * self.A2[s, s]
            )
            y_list.append(y)
            f_list.append(f(y))
            g = (Maxwellian(U) - y_list[s]) / tau_list[s]
            g_list.append(g)
            # print(y.sum())

        # print(len(f_list),len(g_list))
        y = y0 + h * (
            sum([f_list[j] * self.b1[j] for j in range(self.Stage)])
            + sum([(g_list[j]) * self.b2[j] for j in range(self.Stage)])
        )
        return y


class IMEX_RK1(IMEX_RK):
    def __init__(self) -> None:
        b1 = np.array([1])
        A1 = np.array([[0]])
        b2 = np.array([1])
        A2 = np.array([[1]])
        super().__init__(b1, A1, b2, A2)


class IMEX_RK2(IMEX_RK):
    def __init__(self) -> None:
        b1 = np.array([0, 0.5, 0.5])
        A1 = np.array([[0, 0, 0], [0, 0, 0], [0, 1, 0]])
        b2 = np.array([0, 0.5, 0.5])
        A2 = np.array([[0.5, 0, 0], [-0.5, 0.5, 0], [0, 0.5, 0.5]])
        super().__init__(b1, A1, b2, A2)


class IMEX_RK3(IMEX_RK):
    def __init__(self) -> None:
        alpha = 0.24169426078821
        beta = 0.06042356519705
        eta = 0.1291528696059
        b1 = np.array([0, 1 / 6, 1 / 6, 2 / 3])
        A1 = np.array([[0, 0, 0, 0], [0, 0, 0, 0], [0, 1, 0, 0], [0, 1 / 4, 1 / 4, 0]])
        b2 = np.array([0, 1 / 6, 1 / 6, 2 / 3])
        A2 = np.array(
            [
                [alpha, 0, 0, 0],
                [-alpha, alpha, 0, 0],
                [0, 1 - alpha, alpha, 0],
                [beta, eta, 0.5 - alpha - beta - eta, alpha],
            ]
        )
        super().__init__(b1, A1, b2, A2)


#  Stiffly Accurate Scheme, which mean A2[-1,:]=b2
class IMEX_RK1_stiff(IMEX_RK):
    def __init__(self) -> None:
        b1 = np.array([1, 0])
        A1 = np.array([[0, 0], [1, 0]])
        b2 = np.array([0, 1])
        A2 = np.array([[0, 0], [0, 1]])
        super().__init__(b1, A1, b2, A2)


class IMEX_RK2_stiff(IMEX_RK):
    def __init__(self) -> None:
        b1 = np.array([0, 1, 0])
        A1 = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]])
        b2 = np.array([0.5, 0, 0.5])
        A2 = np.array([[0, 0, 0], [0, 1, 0], [0.5, 0.0, 0.5]])
        super().__init__(b1, A1, b2, A2)



import torch
from torch.fft import fftn,ifftn,fftshift
import numpy as np
import torch.nn as nn
import math
from scipy import special
from bte.dvm.distribution import DVDis

def get_potential(omega):
	if(omega==0.5):
		alpha=1.0
	else:
		eta=4.0/(2.0*omega-1.0)+1.0
		alpha=(eta-5.0)/(eta-1.0)
	return alpha

def lgwt(N,a,b):
    x,w=np.polynomial.legendre.leggauss(N)
    x = 0.5*(x + 1)*(b - a) + a
    w = w * 0.5 * (b-a)
    return torch.from_numpy(x),torch.from_numpy(w)

def init_kernel_mode_vector(umax,umin,unum,vmax,vmin,vnum,wmax,wmin,wnum,quad_num,omega=0.81,M=5,dtype=torch.float64):
    PI=math.pi
    alpha=get_potential(omega)
    du,dv,dw=(umax-umin)/(unum-1),(vmax-vmin)/(vnum-1),(wmax-wmin)/(wnum-1)
    supp=math.sqrt(2.0)*2.0*max(umax,vmax,wmax)/(3.0+math.sqrt(2.0))

    fre_vx=torch.linspace(-PI/du,(unum/2-1.0)*2.0*PI/unum/du,unum)
    fre_vy=torch.linspace(-PI/dv,(vnum/2-1.0)*2.0*PI/vnum/dv,vnum)
    fre_vz=torch.linspace(-PI/dw,(wnum/2-1.0)*2.0*PI/wnum/dw,wnum)

    abscissa, gweight=lgwt(quad_num, 0.0, supp)
    theta=PI/M*torch.arange(1,M-1+1)
    theta2=PI/M*torch.arange(1,M+1)

    s=((fre_vx[:,None,None]*torch.sin(theta)[None,:,None]*torch.cos(theta2)[None,None,:])[:,None,None,:,:]+
        (fre_vy[:,None,None]*torch.sin(theta)[None,:,None]*torch.sin(theta2)[None,None,:])[None,:,None,:,:]+
        (fre_vz[:,None,None]*torch.cos(theta)[None,:,None])[None,None,:,:,:]
    ) 

    int_temp= (2 * gweight[...,None,None,None,None,None] * 
                torch.cos( s[None,...] * abscissa[...,None,None,None,None,None] ) * 
                (abscissa[...,None,None,None,None,None]**alpha)).sum(dim=0)
    phi2 = int_temp*torch.sin(theta[None,None,None,:,None]) 

    s=((fre_vx*fre_vx)[:,None,None,None,None]+
        (fre_vy*fre_vy)[None,:,None,None,None]+
        (fre_vz*fre_vz)[None,None,:,None,None] - s*s)
    
    psi2 = torch.zeros((unum,vnum,wnum,M-1,M),dtype=dtype)
    
    so=s.clone()
    s=s.abs()
    s = torch.sqrt(s)
    bel = supp*s
    bessel = torch.from_numpy(special.jv(1,bel.numpy()))

    psi2=2.0*PI*supp*bessel/s
    psi2[so<=0] = PI*supp*supp
    phipsi2=(phi2*psi2).sum(dim=(-1,-2))

    return phi2,psi2,phipsi2


def collision_fft(f_spec, kn_bzm, phi, psi, phipsi)->torch.Tensor:
    unum,vnum,wnum=phi.shape[:3]
    ifft3d=lambda x:ifftn(x,dim=(-3,-2,-1),norm="forward")
    fft3d=lambda x:fftn(x,dim=(-3,-2,-1),norm="backward")

    f_spec=ifft3d(f_spec)
    f_spec=f_spec/(unum*vnum*wnum)
    
    f_spec=fftshift(f_spec,dim=(-3,-2,-1))
    f_temp=0
    M=phi.shape[-1]
    for i in range(1,M-1+1):
        for j in range(1,M+1):
            fc1=f_spec*phi[:,:,:,i-1,j-1]
            fc2=f_spec*psi[:,:,:,i-1,j-1]
            fc11=fft3d(fc1)
            fc22=fft3d(fc2)
            f_temp=f_temp+fc11*fc22
    fc1=f_spec*phipsi
    fc2=f_spec
    fc11=fft3d(fc1)
    fc22=fft3d(fc2)
    f_temp=f_temp-fc11*fc22
    Q = 4.0*np.pi**2/kn_bzm/M**2*f_temp.real
    return Q

def get_collision(v_meta,quad_num=8,omega=0.81,M=5):
    u=v_meta.v_dims[0].v
    v=v_meta.v_dims[1].v
    w=v_meta.v_dims[2].v
    umax=u.max().item()
    umin=u.min().item()
    unum=u.numel()
    vmax=v.max().item()
    vmin=v.min().item()
    vnum=v.numel()
    wmax=w.max().item()
    wmin=w.min().item()
    wnum=w.numel()    
    return init_kernel_mode_vector(umax,umin,unum,vmax,vmin,vnum,wmax,wmin,wnum,quad_num,omega=omega,M=M)

def get_vshape(v_meta):
    u=v_meta.v_dims[0].v
    v=v_meta.v_dims[1].v
    w=v_meta.v_dims[2].v
    unum=u.numel()
    vnum=v.numel()
    wnum=w.numel()  
    return unum,vnum,wnum

class collisioner(nn.Module):
    def __init__(self,v_meta,quad_num,omega=0.81,M=5,device="cpu",dtype=torch.float32):  
        super().__init__()
        phi, psi, phipsi = get_collision(v_meta,quad_num,omega,M)
        self.register_buffer("phi",phi.to(device=device,dtype=dtype))
        self.register_buffer("psi",psi.to(device=device,dtype=dtype))
        self.register_buffer("phipsi",phipsi.to(device=device,dtype=dtype))
        self.f_shape = get_vshape(v_meta)

    def do_collision(self, f:torch.Tensor, kn_bzm=1.0):
        f1=f.reshape(f.shape[:-1]+self.f_shape)
        if isinstance(kn_bzm,torch.Tensor):
            kn_bzm=kn_bzm.reshape_as(f1)
        Q=collision_fft(f1,kn_bzm,self.phi,self.psi,self.phipsi)
        Q=Q.reshape_as(f)
        return Q

class collisioner_bgk(nn.Module):
    def __init__(self,v_meta):
        super().__init__()
        self.v_meta=v_meta

    def do_collision(self, f:torch.Tensor, kn_bzm=1.0):
        F=DVDis(self.v_meta,f)
        return (F.maxwellian()-f)/kn_bzm


import torch
import itertools
import math
import functools 
from typing import Tuple, List

def get_index_table(M:int,dim:int=3):
    """
    get_index_table return two dictionary.
    The first is (int,int,...)->int
    The second is int->(int,int,...)

    Args:
        M (int): order
        dim (int, optional): dimension. Defaults to 3.

    Returns:
        indexNto1dict, index1toNdict
    """
    indexNto1dict={}
    index1toNdict={}
    c=0
    for m in range(M+1):
        for index in itertools.product(*([range(m+1)]*dim)):
            re_index=index[::-1]    
            if sum(re_index)==m:
                indexNto1dict[re_index]=c
                index1toNdict[c]=re_index
                c=c+1
    return indexNto1dict, index1toNdict

def get_index_len(M:int,dim=3):
    """
    get_index_len 

    Args:
        M (int): order.
        dim (int, optional): dimension. Defaults to 3.

    Returns:
        int: (M+dim)*(M+dim-1)*...(M+1)/(dim!)
    """
    return functools.reduce(lambda x,y:x*y,[M+i+1 for i in range(dim)])//math.factorial(dim)

class index_table():
    def __init__(self,M:int,dim:int=3):
        indexNto1dict, index1toNdict = get_index_table(M,dim)
        self.iNto1 = indexNto1dict
        self.i1toN = index1toNdict
        self.ORDER = M
        self.DIM = dim
        self.len = len(indexNto1dict)
    def get_order(self,index1,dim=0):
        assert dim<=self.DIM
        return [self.i1toN[ind][dim] for ind in index1]
        
def get_change_index(indt:index_table,index_base:List[int],index_tuple_add):
    outputs=[]
    assert len(indt.i1toN[0])==len(index_tuple_add)
    for index in index_base:
        index_tuple=indt.i1toN[index]
        index_change=tuple([index_tuple[i]+index_tuple_add[i] for i in range(len(index_tuple_add))])
        outputs.append(indt.iNto1.get(index_change,None))
    return outputs

def get_change_index_projection(indt:index_table,index_base:List[int],index_tuple_add):
    tmp1=get_change_index(indt,index_base,index_tuple_add)    
    origin=[]
    goal=[]
    for i in index_base:
        if tmp1[i] is not None:
            origin.append(i)
            goal.append(tmp1[i])
    return torch.tensor(origin),torch.tensor(goal)

class index_table_expand(index_table):
    def __init__(self,M:int,dim:int=3):
        super().__init__(M, dim)
        length=len(self.iNto1)
        DIM=dim
        self.d1=[]
        self.u1=[]
        self.d2=[]
        v1=[]
        v2=[]
        m1=[]
        for i in range(DIM):
            d1=[0,]*DIM
            d1[i]=-1
            id1=get_change_index_projection(self,list(range(length)),tuple(d1))
            self.d1.append(id1)

            d2=[0,]*DIM
            d2[i]=-2
            id2=get_change_index_projection(self,list(range(length)),tuple(d2))
            self.d2.append(id2)

            u1=[0,]*DIM
            u1[i]=1
            iu1=get_change_index_projection(self,list(range(length)),tuple(u1))
            self.u1.append(iu1)
            v1.append(self.iNto1[tuple(u1)])
          
            u2=[0,]*DIM
            u2[i]=2
            v2.append(self.iNto1[tuple(u2)])
            
            m1.append(torch.tensor(self.get_order(self.u1[i][0].tolist(),dim=i)) + 1)
                      
        self.v1=torch.tensor(v1)
        self.v2=torch.tensor(v2)
        self.m1=m1
        
        
        
class index_table_global_():
    def __init__(self):
        self.cache={}
    def get(self,ORDER,DIM):
        if (ORDER,DIM) not in self.cache:
            self.cache[(ORDER,DIM)]=index_table_expand(ORDER,DIM)
        return self.cache[(ORDER,DIM)]
    
index_tables = index_table_global_()


import numpy as np
from collections.abc import Callable

class RK:
    r"""Explicit Runge-Kutta Method

    Create a runge-kutta method to integrate :math:`u_t=f(t,u)`

    .. math::
        \begin{aligned} y_{n+1}&=y_{n}+h\sum _{i=1}^{s}b_{i}k_{i}\\k_{1}&=f(t_{n},y_{n}),\\k_{2}&=f(t_{n}+c_{2}h,y_{n}+h(a_{21}k_{1})),\\k_{3}&=f(t_{n}+c_{3}h,y_{n}+h(a_{31}k_{1}+a_{32}k_{2})),\\&\ \ \vdots \\k_{s}&=f(t_{n}+c_{s}h,y_{n}+h(a_{s1}k_{1}+a_{s2}k_{2}+\cdots +a_{s,s-1}k_{s-1})).
        \end{aligned}

    Args:
        c (list): runge-kutta coef
        b (list): runge-kutta coef
        A (list): runge-kutta coef

    Butcher tableau
    
    .. math::
        \begin{array}{c|c}
        c & A \\ \hline
          & b
        \end{array}

    """

    def __init__(self, c: list, b: list, A: list) -> None:
        self.Stage = len(c)
        if len(b) != self.Stage or A.shape[0] != A.shape[1] or A.shape[0] != self.Stage:
            raise ValueError("Invalid Butcher tableau")
        self.c = c
        self.b = b
        self.A = A

    def __call__(self, f: Callable, t0: float, u0: float, h: float, f0=None):
        """ single step of runge-kutta

        u_t = f(u,t)
        
        f_0=f(u0,t0)

        Args:
            f (Callable): [description]
            t0 (float): [description]
            u0 (float): [description]
            h (float): [description]
            f0 ([type], optional): [description]. Defaults to None.

        Returns:
            [type]: [description]
        """
        k_list = []
        for i in range(self.Stage):
            if i == 0 and f0 is not None and self.A[0, 0] == 0:
                k_list.append(f0)
            else:
                k_list.append(
                    f(
                        t0 + self.c[i] * h,
                        u0 + h * sum([self.A[i, j] * k_list[j] for j in range(i)]),
                    )
                )
        u = u0 + h * sum([k_list[j] * self.b[j] for j in range(self.Stage)])
        return u


class RK1(RK):
    r""" 1st order Runge-Kutta

    Butcher tableau:
    
    .. math::
        \begin{array}{c|c}
        0 & 0 \\ \hline
          & 1
        \end{array}
    """

    def __init__(self) -> None:
        c = np.array([0])
        b = np.array([1])
        A = np.array([[0]])
        super().__init__(c, b, A)


class RK2(RK):
    r""" 2nd order Runge-Kutta

    Butcher tableau:
    
    .. math::
        \begin{array}{c|cc}
        0 & 0 & 0\\ 
        \alpha & \alpha & 0 \\ \hline
          & (1 - \frac{1}{2\alpha}) & \frac{1}{2\alpha}
        \end{array}

    Args:
        alpha (float, optional): Defaults to 1/2.
    """

    def __init__(self, alpha:float=1/2) -> None:
        c = np.array([0, alpha])
        b = np.array([1 - 1 / (2 * alpha), 1 / (2 * alpha)])
        A = np.array([[0, 0], [alpha, 0]])
        super().__init__(c, b, A)


class RK3_SSP(RK):
    def __init__(self) -> None:
        c = np.array([0, 1, 0.5])
        b = np.array([1 / 6, 1 / 6, 2 / 3])
        A = np.array([[0, 0, 0], [1, 0, 0], [1 / 4, 1 / 4, 0]])
        super().__init__(c, b, A)


class RK4(RK):
    r""" 4nd order Runge-Kutta

    Butcher tableau:

    .. math::
        \begin{array}{c|cc}
        0 &  & & &\\ 
        \frac{1}{2} & \frac{1}{2} & & & \\ 
        \frac{1}{2} & 0 & \frac{1}{2} & & \\
        1 & 0 & 0 & 1 & \\ \hline
        & \frac{1}{6} & \frac{1}{3} & \frac{1}{3} & \frac{1}{6}
        \end{array}

    """

    def __init__(self) -> None:
        c = np.array([0, 0.5, 0.5, 1])
        b = np.array([1 / 6, 1 / 3, 1 / 3, 1 / 6])
        A = np.array([[0, 0, 0, 0], [1 / 2, 0, 0, 0], [0, 1 / 2, 0, 0], [0, 0, 1, 0]])
        super().__init__(c, b, A)


class RK4_2(RK):
    def __init__(self) -> None:
        c = np.array([0, 1 / 3, 2 / 3, 1])
        b = np.array([1 / 8, 3 / 8, 3 / 8, 1 / 8])
        A = np.array([[0, 0, 0, 0], [1 / 3, 0, 0, 0], [-1 / 3, 1, 0, 0], [1, -1, 1, 0]])
        super().__init__(c, b, A)


if __name__ == "__main__":
    a = RK4_2()
    f = lambda t, x: x
    u = 1
    for i in range(10):
        u = a(f, 0, u, 1 / 10)
    print("error is:", u - np.exp(1))


import torch
import torch.nn as nn

from scipy.special import roots_hermitenorm
from bte.grad.distribution import HermiteDisND

from bte.utils.weno5 import *
from bte.utils.limiter import *
from bte.grad.distribution import HermiteDis
from bte.grad.distribution import HermitedistributionBase as HDISBASE
from bte.utils.indexs import *
from bte.grad.collision_grad import collision_f, load_collision_kernel
Tensor = torch.Tensor


def MaxRootOfHermitePolynomial(N: int) -> float:
    return roots_hermitenorm(N)[0][-1]


def simpson(l, m, r):
    return (l + 4 * m + r) / 6


def linear_reconstruction(
    dis: HDISBASE, disL: HDISBASE, disR: HDISBASE, limiter=limiter_mc
):
    theta = torch.zeros_like(dis.coef)
    theta1 = dis.coef - disL.coef
    theta2 = disR.coef - dis.coef
    mask = torch.abs(theta1 * theta2) > 1e-12
    theta[mask] = (theta1 / (theta2 + 1e-12))[mask]

    theta = limiter(theta)
    RecL = dis.coef - theta / 2 * theta2
    RecR = dis.coef + theta / 2 * theta2

    return dis.new(dis.u, dis.theta, RecL), dis.new(dis.u, dis.theta, RecR)


class Equation_HermiteBased(nn.Module):
    def __init__(self, M=5, bdc="circular"):
        super().__init__()
        self.M = M
        self.MRoH = MaxRootOfHermitePolynomial(self.M)
        self.veloc = self.Grad_Vec_LLF
        self.flux_solver = "LLF"
        self.closure = self.Grad_Closure
        self.Reconstruction = self.Reconstruction_linear
        self.ode_step = self.Heun
        self.PROJ_STEPS = 1
        self.bdc = bdc
        self.split_order = 1
        self.collision = self.collision_BGK
        self.bgk_simple_kn ="simple2"
        self.max_dt = 1.0
        self.hme = False
        self.nrxx = False
        self.CFL = 0.45
        self.chi = 0
        self.padding_order = 2
        self.no_closure=False
        
    def set_collision(self, colltype:str):
        assert colltype=="BGK" or colltype=="Maxwellian"
        if colltype=="BGK":
            self.collision=self.collision_BGK
        elif colltype=="Maxwellian":
            self.set_QKernel()
            self.collision=self.collision_Kernel
        else:
            pass
        return 

    def set_QKernel(self):
        self.register_buffer("QKernel",load_collision_kernel(order=self.M))

    def __str__(self) -> str:
        base_str = super().__str__()
        infomation = f"""
====
Equation_HermiteBased
----
flux_solver:{self.flux_solver}
split_order:{self.split_order}
PROJ_STEPS:{self.PROJ_STEPS}
bgk_simple_kn:{self.bgk_simple_kn}
===="""
        return base_str + "\n" + infomation

    def set_order(self, space_order, time_order):
        self.set_space_order(space_order)
        self.set_time_order(time_order)

    def set_space_order(self, order):
        if order == 1:
            self.Reconstruction = self.Reconstruction_constant
            self.padding_order=1
        elif order == 2:
            self.Reconstruction = self.Reconstruction_linear
            self.padding_order=2
        else:
            raise ValueError

    def set_time_order(self, order):
        if order == 1:
            self.ode_step = self.Euler
        elif order == 2:
            self.ode_step = self.Heun
        else:
            raise ValueError

    def set_initial(self, kn: torch.Tensor, f0: HDISBASE, dx: float):
        # assert kn.shape[-1] == 1
        # assert f0.coef.shape[-1] == self.M
        # assert kn.shape == f0.u.shape
        self.kn = kn
        self.f0 = f0
        self.dx = dx
        if self.bdc=="constant":
            self.left_BC=f0.new(f0.u[...,0:1,:],f0.theta[...,0:1,:],f0.coef[...,0:1,:])
            self.right_BC=f0.new(f0.u[...,-1:,:],f0.theta[...,-1:,:],f0.coef[...,-1:,:])

    def Grad_Vec(self, dis: HDISBASE, Kn=None):
        u0 = dis.u
        s0 = torch.sqrt(dis.theta)
        return u0 + self.MRoH * s0, u0 - self.MRoH * s0

    def Grad_Vec_LLF(self, dis: HDISBASE, Kn=None):
        u0 = dis.u
        s0 = torch.sqrt(dis.theta)
        return torch.maximum(
            torch.abs(u0 + self.MRoH * s0), torch.abs(u0 - self.MRoH * s0)
        )

    def Grad_Closure(self, dis: HDISBASE, Kn=None):
        if isinstance(dis,HermiteDisND):
            dl=index_tables.get(dis.indt.ORDER+1,dis.indt.DIM).len-dis.indt.len
            ex = torch.zeros(dis.u.shape[:-1]+(dl,), device=dis.coef.device)
        else:
            ex = torch.zeros_like(dis.u)

        return ex

    def NRxx_Closure(self, dis: HDISBASE, Kn: Tensor) -> Tensor:
        disM = dis
        disL = disM.roll(-1)
        disR = disM.roll(1)
        disLP = disL.Project_2B(disM, STEP=self.PROJ_STEPS)
        disRP = disR.Project_2B(disM, STEP=self.PROJ_STEPS)
        theta = disM.theta
        closure = (
            -theta * Kn * (disRP.coef[..., -1:] - disLP.coef[..., -1:]) / (2 * self.dx)
        )
        return closure

    def getTau(self, dis, Kn):
        if self.bgk_simple_kn=="simple":
            tau = self.kn 
        elif self.bgk_simple_kn=="simple2":
            tau = Kn / dis.density()
        elif self.bgk_simple_kn=="mu_ref":
            tau =  self.mu_ref*2/(dis.theta*2)**(1-self.omega)/dis.density()
        else:
            raise ValueError
        return tau

    def collision_BGK(self, dis: HDISBASE, dt, Kn):
        # if torch.tensor(dt).numel() == 1:
        #     dt = torch.tensor(dt, device=dis.u.device).repeat(dis.u.shape[0])
        dt = dt.reshape((dis.u.shape[0], 1, 1))
        # if self.bgk_simple_kn:
        #     ceo = Kn
        # else:
        #     #ceo = Kn / dis.coef[..., 0:1]
        #     ceo = self.mu_ref*2/(dis.theta*2)**(1-self.omega)/dis.density()
        ceo=self.getTau(dis,Kn)
        tmp = dis.coef.clone()
        tmp[..., 1:] = dis.coef[..., 1:] * torch.exp(-dt / ceo)
        return dis.new(dis.u, dis.theta, tmp)

    def collision_Kernel(self, dis:HDISBASE, dt, Kn):
        dt = dt.reshape((dis.u.shape[0], 1, 1))
        if self.bgk_simple_kn:
            ceo = Kn
        else:
            ceo = Kn / dis.coef[..., 0:1]
        tmp = dis.coef.clone()
        Q = collision_f(self.QKernel, tmp)
        tmp = tmp + dt / ceo * Q
        return dis.new(dis.u, dis.theta, tmp)

    def get_tstep_cell(self, dis: HDISBASE):
        """
        因为一般是在碰撞后求解下一步的时间步长，所以直接读取，
        若非如此，需要 _,c,theta=get_pcs(dis)
        """
        c, theta = dis.u[..., 0], dis.theta[..., 0]
        return torch.abs(c) + self.MRoH * torch.sqrt(theta)

    def get_tstep(self, dislist: HDISBASE, dx):
        z = self.get_tstep_cell(dislist)
        tstep = self.CFL * dx / z.max(dim=1)[0]
        if self.nrxx:
            tstep = self.CFL * dx * dx / z.max(dim=1)[0]
        return tstep

    def LLF_helper(
        self,
        disL: HDISBASE,
        disR: HDISBASE,
        fluxL: HDISBASE,
        fluxR: HDISBASE,
        lam,
    ):
        #print(fluxL.coef.shape,lam.shape)
        tmp = 0.5 * (fluxL.coef + fluxR.coef) + 0.5 * lam * (disL.coef - disR.coef)
        return disL.new(disL.u, disL.theta, tmp)

    def HLL_helper(
        self,
        disL: HDISBASE,
        disR: HDISBASE,
        fluxL: HDISBASE,
        fluxR: HDISBASE,
        lambda_L,
        lambda_R,
    ) -> HDISBASE:
        k0 = (torch.abs(lambda_L) - torch.abs(lambda_R)) / (lambda_L - lambda_R)
        k1 = (torch.abs(lambda_L) * lambda_R - torch.abs(lambda_R) * lambda_L) / (
            lambda_L - lambda_R
        )
        tmp = (
            0.5 * (fluxL.coef + fluxR.coef)
            + 0.5 * k0 * (fluxL.coef - fluxR.coef)
            - 0.5 * k1 * (disL.coef - disR.coef)
        )
        return disL.new(disL.u, disL.theta, tmp)

    def flux_helper(
        self,
        disL: HDISBASE,
        disR: HDISBASE,
        fluxL: HDISBASE,
        fluxR: HDISBASE,
        Kn,
        disLO: HDISBASE,
        disRO: HDISBASE,
        lam=0,
    ) -> HDISBASE:
        if self.flux_solver == "HLL":
            vLmax, vLmin = self.veloc(disLO, Kn)
            vRmax, vRmin = self.veloc(disRO, Kn)

            lambda_min = torch.minimum(vLmin, vRmin)
            lambda_max = torch.maximum(vLmax, vRmax)

            R_Flux = self.HLL_helper(disL, disR, fluxL, fluxR, lambda_min, lambda_max)
        elif self.flux_solver == "LF":
            lam = lam.reshape((disL.u.shape[0], 1, 1))
            R_Flux = self.LLF_helper(disL, disR, fluxL, fluxR, lam)
        elif self.flux_solver == "LLF":
            vLmax = self.veloc(disLO, Kn)
            vRmax = self.veloc(disRO, Kn)
            lambda_max = torch.max(vLmax, vRmax)[...,0:1]
            R_Flux = self.LLF_helper(disL, disR, fluxL, fluxR, lambda_max)
        else:
            raise NotImplementedError

        return R_Flux

    def get_flux_hme_helper(self, disL: HDISBASE, disR: HDISBASE):
        f_l = disL.coef[..., -1]
        f_r = disR.coef[..., -1]
        f_l1 = disL.coef[..., -2]
        f_r1 = disR.coef[..., -2]
        rho_l = disL.coef[..., 0]
        rho_r = disR.coef[..., 0]
        diff_u = disR.u[..., 0] - disL.u[..., 0]
        diff_theta = disR.theta[..., 0] - disL.theta[..., 0]
        tmp = -(
            simpson(f_l, (f_l / rho_l + f_r / rho_r) * (rho_l + rho_r) / 4, f_r)
            * diff_u
            + 0.5
            * simpson(f_l1, (f_l1 / rho_l + f_r1 / rho_r) * (rho_l + rho_r) / 4, f_r1)
            * diff_theta
        )
        return tmp * (self.M)

    def get_flux_hme(self, dis: HDISBASE, dx, Kn):
        disM = dis
        u, theta, coefM = dis.as_tuple()
        disL = disM.roll(-1)
        disR = disM.roll(1)
        tmp_L = torch.zeros_like(disM.coef)
        tmp_R = torch.zeros_like(disM.coef)

        coef0 = torch.zeros_like(disM.coef)

        tmp_L[..., -1] = self.get_flux_hme_helper(disL, disM)
        tmp_R[..., -1] = self.get_flux_hme_helper(disM, disR)

        dis_zero = (u, theta, coef0)
        fluxLC = self.flux_helper(
            dis_zero, dis_zero, (u, theta, -tmp_L), dis_zero, Kn, disL, disM
        )
        fluxRC = self.flux_helper(
            dis_zero, dis_zero, dis_zero, (u, theta, tmp_R), Kn, disM, disR
        )
        sum_flux = fluxRC - fluxLC

        return sum_flux

    def padding_periodic_dis(self, f: HDISBASE):
        disP = f.new(
            self.padding_periodic_single(f.u),
            self.padding_periodic_single(f.theta),
            self.padding_periodic_single(f.coef),
        )
        return disP

    def padding_replicate_dis(self, f: HDISBASE):
        disP = f.new(
            self.padding_replicate_single(f.u),
            self.padding_replicate_single(f.theta),
            self.padding_replicate_single(f.coef),
        )
        return disP

    def padding_constant_dis(self, f: HDISBASE):
        u = f.u.transpose(-1, -2)
        ubc = F.pad(
            u, pad=(self.padding_order, self.padding_order), mode="constant"
        ).transpose(-1, -2)
        ubc[..., 0 : self.padding_order, :] = self.left_BC.u
        ubc[..., -self.padding_order :, :] = self.right_BC.u

        T = f.theta.transpose(-1, -2)
        Tbc = F.pad(
            T, pad=(self.padding_order, self.padding_order), mode="constant"
        ).transpose(-1, -2)
        Tbc[..., 0 : self.padding_order, :] = self.left_BC.theta
        Tbc[..., -self.padding_order :, :] = self.right_BC.theta

        coef = f.coef.transpose(-1, -2)
        coefbc = F.pad(
            coef, pad=(self.padding_order, self.padding_order), mode="constant"
        ).transpose(-1, -2)
        coefbc[..., 0 : self.padding_order, :] = self.left_BC.coef
        coefbc[..., -self.padding_order :, :] = self.right_BC.coef

        disP = f.new(
            ubc,
            Tbc,
            coefbc
        )
        
        return disP

    def padding_reflect_dis(self, f: HDISBASE):
        disP = f.new(
            self.padding_reflect_single_u(f.u),
            self.padding_reflect_single(f.theta),
            self.padding_reflect_single(f.coef),
        )
        return disP

    def padding_periodic_single(self, f: torch.Tensor):
        f = f.transpose(-1, -2)
        return F.pad(
            f, pad=(self.padding_order, self.padding_order), mode="circular"
        ).transpose(-1, -2)

    def padding_replicate_single(self, f: torch.Tensor):
        f = f.transpose(-1, -2)
        return F.pad(
            f, pad=(self.padding_order, self.padding_order), mode="replicate"
        ).transpose(-1, -2)

    def padding_reflect_single(self, f: torch.Tensor):
        fp = torch.zeros(
            *f.shape[0:-2],
            f.shape[-2] + 2 * self.padding_order,
            f.shape[-1],
            device=f.device,
        )
        fp[..., self.padding_order : -self.padding_order, :] = f
        for i in range(self.padding_order):
            fp[..., self.padding_order - i - 1, 0::2] = fp[
                ..., self.padding_order + i, 0::2
            ]
            fp[..., -self.padding_order + i, 0::2] = fp[
                ..., -self.padding_order - i - 1, 0::2
            ]
            fp[..., self.padding_order - i - 1, 1::2] = -fp[
                ..., self.padding_order + i, 1::2
            ]
            fp[..., -self.padding_order + i, 1::2] = -fp[
                ..., -self.padding_order - i - 1, 1::2
            ]
        return fp

    def padding_reflect_single_u(self, f: torch.Tensor):
        fp = torch.zeros(
            *f.shape[0:-2],
            f.shape[-2] + 2 * self.padding_order,
            f.shape[-1],
            device=f.device,
        )
        fp[..., self.padding_order : -self.padding_order, :] = f
        for i in range(self.padding_order):
            fp[..., self.padding_order - i - 1, 0::2] = -fp[
                ..., self.padding_order + i, 0::2
            ]
            fp[..., -self.padding_order + i, 0::2] = -fp[
                ..., -self.padding_order - i - 1, 0::2
            ]
        return fp

    def padding_constant_single(self, f: torch.Tensor):
        f = f.transpose(-1, -2)
        fbc = F.pad(
            f, pad=(self.padding_order, self.padding_order), mode="constant"
        ).transpose(-1, -2)
        fbc[..., 0 : self.padding_order, :] = self.left_BC
        fbc[..., -self.padding_order :, :] = self.right_BC
        return fbc

    def padding_dis(self, f: torch.Tensor):
        if self.bdc == "circular":
            return self.padding_periodic_dis(f)
        elif self.bdc == "constant":
            return self.padding_constant_dis(f)
        elif self.bdc == "cauchy":
            return self.padding_replicate_dis(f)
        elif self.bdc == "maxwellian":
            return self.padding_reflect_dis(f)
        else:
            return

    def Reconstruction_constant(self, dis: HDISBASE):
        #print("dis.u.shape:", dis.u.shape)
        disL = dis.new(
            dis.u[..., :-2, :], dis.theta[..., :-2, :], dis.coef[..., :-2, :]
        )
        disM = dis.new(
            dis.u[..., 1:-1, :], dis.theta[..., 1:-1, :], dis.coef[..., 1:-1, :]
        )
        disR = dis.new(
            dis.u[..., 2:, :], dis.theta[..., 2:, :], dis.coef[..., 2:, :]
        )
        disLexP = disL.Project_2B(disM, STEP=self.PROJ_STEPS)
        disRexP = disR.Project_2B(disM, STEP=self.PROJ_STEPS)
        return disLexP, disM, disM, disRexP, disL, disR

    def Reconstruction_linear(self, dis: HDISBASE):
        disLL = dis.new(
            dis.u[..., :-4, :], dis.theta[..., :-4, :], dis.coef[..., :-4, :]
        ).clone()
        disL = dis.new(
            dis.u[..., 1:-3, :], dis.theta[..., 1:-3, :], dis.coef[..., 1:-3, :]
        ).clone()
        disM = dis.new(
            dis.u[..., 2:-2, :], dis.theta[..., 2:-2, :], dis.coef[..., 2:-2, :]
        ).clone()
        disR = dis.new(
            dis.u[..., 3:-1, :], dis.theta[..., 3:-1, :], dis.coef[..., 3:-1, :]
        ).clone()
        disRR = dis.new(
            dis.u[..., 4:, :], dis.theta[..., 4:, :], dis.coef[..., 4:, :]
        ).clone()

        disLP = disL.Project_2B(disM, STEP=self.PROJ_STEPS)
        disRP = disR.Project_2B(disM, STEP=self.PROJ_STEPS)
        disLLP = disLL.Project_2B(disM, STEP=self.PROJ_STEPS)
        disRRP = disRR.Project_2B(disM, STEP=self.PROJ_STEPS)

        _, disLP_R = linear_reconstruction(disLP, disLLP, disM)
        disRP_L, _ = linear_reconstruction(disRP, disM, disRRP)
        disMP_L, disMP_R = linear_reconstruction(disM, disLP, disRP)

        return disLP_R, disMP_L, disMP_R, disRP_L, disL, disR

    def Reconstruction_WENO(self, disM: HDISBASE):
        raise NotImplementedError
        # disLL = disM.roll(-2)
        # disL = disM.roll(-1)
        # disR = disM.roll(1)
        # disRR = disM.roll(2)

        # disLP = disL.Project_AB(disM, STEP=self.PROJ_STEPS)
        # disRP = disR.Project_AB(disM, STEP=self.PROJ_STEPS)
        # disLLP = disLL.Project_AB(disM, STEP=self.PROJ_STEPS)
        # disRRP = disRR.Project_AB(disM, STEP=self.PROJ_STEPS)

        # disMP_Lm, disMP_Rm = weno5_LR(
        #     disLLP.coef, disLP.coef, disM.coef, disRP.coef, disRRP.coef
        # )
        # disMP_L = HermiteDis(disM.u, disM.theta, disMP_Lm)
        # disMP_R = HermiteDis(disM.u, disM.theta, disMP_Rm)
        # Recon_disLR = disMP_R.roll(-1)
        # Recon_disLR = Recon_disLR.Project_2B(disM, STEP=self.PROJ_STEPS)

        # Recon_disRL = disMP_L.roll(1)
        # Recon_disRL = Recon_disRL.Project_2B(disM, STEP=self.PROJ_STEPS)

        # return Recon_disLR, disMP_L, disMP_R, Recon_disRL, disL, disR

    def maxwellian_bdc_fix(self, disLP_R, disMP_L, disMP_R, disRP_L, disLO, disRO):
        # disLP_R,disMP_L,disMP_R,disRP_L,disLO,disRO
        # 左cell的右边界，中间cell的左边界，中间cell的右边界，右边cell的左边界。左边cell的原始值，右边cell的原始值。
        # 要改的是左cell的右边界、右cell的左边界。要改idx=0和idx=-1的cell
        if self.chi == 0:
            disLP_R_u = disLP_R.u.clone()
            disMP_L_u = disMP_L.u.clone()
            disMP_R_u = disMP_R.u.clone()
            disRP_L_u = disRP_L.u.clone()

            disLP_R_theta = disLP_R.theta.clone()
            disMP_L_theta = disMP_L.theta.clone()
            disMP_R_theta = disMP_R.theta.clone()
            disRP_L_theta = disRP_L.theta.clone()

            disLP_R_coef = disLP_R.coef.clone()
            disMP_L_coef = disMP_L.coef.clone()
            disMP_R_coef = disMP_R.coef.clone()
            disRP_L_coef = disRP_L.coef.clone()

            # 修正disLP_R
            disLP_R_3_theta = disMP_L_theta[:, 0, :]
            disLP_R_3_coef = disMP_L_coef[:, 0, :]
            disLP_R_3_coef[:, 1::2] = -disMP_L_coef[:, 0, 1::2]
            disLP_R_u[:, 0, :] = -disMP_L_u[:, 0, :]
            disLP_R_theta[:, 0, :] = disLP_R_3_theta
            disLP_R_coef[:, 0, :] = disLP_R_3_coef

            # 修正disRP_L
            disRP_L_3_theta = disMP_R_theta[:, -1, :]
            disRP_L_3_coef = disMP_R_coef[:, -1, :]
            disRP_L_3_coef[:, 1::2] = -disMP_R_coef[:, -1, 1::2]
            disRP_L_u[:, -1, :] = -disMP_R_u[:, -1, :]
            disRP_L_theta[:, -1, :] = disRP_L_3_theta
            disRP_L_coef[:, -1, :] = disRP_L_3_coef

            disLP_R = disLO.new(disLP_R_u, disLP_R_theta, disLP_R_coef)
            disRP_L = disLO.new(disRP_L_u, disRP_L_theta, disRP_L_coef)

            return disLP_R, disMP_L, disMP_R, disRP_L, disLO, disRO
        else:
            disLP_R_u = disLP_R.u.clone()
            disMP_L_u = disMP_L.u.clone()
            disMP_R_u = disMP_R.u.clone()
            disRP_L_u = disRP_L.u.clone()

            disLP_R_theta = disLP_R.theta.clone()
            disMP_L_theta = disMP_L.theta.clone()
            disMP_R_theta = disMP_R.theta.clone()
            disRP_L_theta = disRP_L.theta.clone()

            disLP_R_coef = disLP_R.coef.clone()
            disMP_L_coef = disMP_L.coef.clone()
            disMP_R_coef = disMP_R.coef.clone()
            disRP_L_coef = disRP_L.coef.clone()

            # 修正disLP_R
            disLP_R_3_theta = disMP_L_theta[:, 0, :]
            disLP_R_3_coef = disMP_L_coef[:, 0, :]
            disLP_R_3_coef[:, 1::2] = -disMP_L_coef[:, 0, 1::2]
            disLP_R_u[:, 0, :] = -disMP_L_u[:, 0, :]
            disLP_R_theta[:, 0, :] = disLP_R_3_theta
            disLP_R_coef[:, 0, :] = disLP_R_3_coef

            # 修正disRP_L
            disRP_L_3_theta = disMP_R_theta[:, -1, :]
            disRP_L_3_coef = disMP_R_coef[:, -1, :]
            disRP_L_3_coef[:, 1::2] = -disMP_R_coef[:, -1, 1::2]
            disRP_L_u[:, -1, :] = -disMP_R_u[:, -1, :]
            disRP_L_theta[:, -1, :] = disRP_L_3_theta
            disRP_L_coef[:, -1, :] = disRP_L_3_coef

            disLP_R = disLO.new(disLP_R_u, disLP_R_theta, disLP_R_coef)
            disRP_L = disLO.new(disRP_L_u, disRP_L_theta, disRP_L_coef)

            return disLP_R, disMP_L, disMP_R, disRP_L, disLO, disRO

    def get_convection_flux(self, dis: HDISBASE, dx, Kn):
        # uM, thetaM, coefM = dis.as_tuple()
        # coefMex = torch.cat((coefM, self.closure(dis, Kn)), -1)
        # if isinstance(dis,HermiteDisND):
        #     disMex = dis.new(uM, thetaM, coefMex, 
        #             index_tables.get(dis.indt.ORDER+1,dis.indt.DIM))
        # else:
        #     disMex = dis.new(uM, thetaM, coefMex)
        # disMexPad = self.padding_dis(disMex)

        disP = self.padding_dis(dis)
        uP, thetaP, coefP = disP.as_tuple()
        coefPex = torch.cat((coefP, self.closure(disP, Kn)), -1)
        if isinstance(dis,HermiteDisND):
            disMexPad = dis.new(uP, thetaP, coefPex, 
                    index_tables.get(dis.indt.ORDER+1,dis.indt.DIM))
        else:
            disMexPad = dis.new(uP, thetaP, coefPex)
        

        disLP_R, disMP_L, disMP_R, disRP_L, disLO, disRO = self.Reconstruction(
            disMexPad
        )

        # 在这里对maxwellian边界处的各阶矩进行修正。
        if self.bdc == "maxwellian":
            disLP_R, disMP_L, disMP_R, disRP_L, disLO, disRO = self.maxwellian_bdc_fix(
                disLP_R, disMP_L, disMP_R, disRP_L, disLO, disRO
            )
        # print("LP_R",disLP_R.u[0,0,:], disLP_R.theta[0,0,:], disLP_R.coef[0,0,:])
        # print("MP_L",disMP_L.u[0,0,:], disMP_L.theta[0,0,:], disMP_L.coef[0,0,:])

        #         print("RP_L",disRP_L.u[0,-1,:], disRP_L.theta[0,-1,:], disRP_L.coef[0,-1,:])
        #         print("MP_R",disMP_R.u[0,-1,:], disMP_R.theta[0,-1,:], disMP_R.coef[0,-1,:])

        Mulvec_LR = disLP_R.mulvec().down()
        Mulvec_ML = disMP_L.mulvec().down()
        Mulvec_MR = disMP_R.mulvec().down()
        Mulvec_RL = disRP_L.mulvec().down()

        #         print("Mulvec",1e2*Mulvec_RL.rho[0,-1,0],1e2*Mulvec_MR.rho[0,-1,0])
        #         print("dis",1e5*disRP_L.rho[0,-1,0],1e5*disMP_R.rho[0,-1,0])
        #         print(1e5*disRO.u[0,-1,0],1e5*disMex.u[0,-1,0])
        #         print(1e5*disRO.theta[0,-1,0],1e5*disMex.theta[0,-1,0])
        #         print("===")
        fluxLC = self.flux_helper(
            disLP_R.down(),
            disMP_L.down(),
            Mulvec_LR,
            Mulvec_ML,
            Kn,
            disLO.down(),
            dis,
        )
        # print(fluxLC.rho[0,0,0])
        fluxRC = self.flux_helper(
            disMP_R.down(),
            disRP_L.down(),
            Mulvec_MR,
            Mulvec_RL,
            Kn,
            dis,
            disRO.down(),
        )
        # print(100000*fluxLC.rho[0,0,0])
        # print(100000*fluxRC.rho[0,-1,0])

        sum_flux = fluxRC - fluxLC

        if self.hme:
            disL = disLO.down()
            disR = disRO.down()
            tmp_L = torch.zeros_like(dis.coef)
            tmp_R = torch.zeros_like(dis.coef)
            coef0 = torch.zeros_like(dis.coef)

            tmp_L[..., -1] = self.get_flux_hme_helper(disL, dis)
            tmp_R[..., -1] = self.get_flux_hme_helper(dis, disR)

            u, theta, coefM = dis.as_tuple()
            dis_zero = dis.new(u, theta, coef0)
            fluxLC = self.flux_helper(
                dis_zero, dis_zero, (u, theta, -tmp_L), dis_zero, Kn, disL, dis
            )  # Rn+
            fluxRC = self.flux_helper(
                dis_zero, dis_zero, dis_zero, (u, theta, tmp_R), Kn, dis, disR
            )  # Rn-
            hme_flux = fluxRC - fluxLC
            sum_flux = sum_flux + hme_flux
        return sum_flux

    def get_flux(self, dis, dx, Kn):
        sum_flux = self.get_convection_flux(dis, dx, Kn) * (1 / dx)
        return sum_flux

    def Euler(self, dis: HDISBASE, dt, dx, Kn):
        if self.split_order == 2:
            dis = self.collision(dis, dt / 2, Kn)
        flux = -dt.reshape((dis.u.shape[0], 1, 1)) * self.get_flux(dis, dx, Kn)
        dis = self.apply_flux(dis, flux)
        dis = dis.Project_to_STD(STEP=self.PROJ_STEPS)
        if self.split_order == 1:
            dis = self.collision(dis, dt, Kn)
        elif self.split_order == 2:
            dis = self.collision(dis, dt / 2, Kn)
        else:
            raise ValueError("split_order must be 1 or 2.")
        return dis

    def Heun(self, dis, dt, dx, Kn):
        if self.split_order == 2:
            dis = self.collision(dis, dt / 2, Kn)

        flux = -dt.reshape((dis.u.shape[0], 1, 1)) * self.get_flux(dis, dx, Kn)
        dis0 = dis.new(dis.u, dis.theta, dis.coef.clone())
        dis1 = self.apply_flux(dis, flux)
        dis1 = dis1.Project_to_STD(STEP=self.PROJ_STEPS)

        flux2 = -dt.reshape((dis.u.shape[0], 1, 1)) * self.get_flux(dis1, dx, Kn)
        flux2 = flux2.Project_2B(dis0, STEP=self.PROJ_STEPS)

        sumflux = 0.5 * flux + 0.5 * flux2
        dis = self.apply_flux(dis0, sumflux)
        dis = dis.Project_to_STD(STEP=self.PROJ_STEPS)

        if self.split_order == 1:
            dis = self.collision(dis, dt, Kn)
        elif self.split_order == 2:
            dis = self.collision(dis, dt / 2, Kn)
        else:
            raise ValueError("split_order must be 1 or 2.")
        return dis

    def apply_flux(self, dis: HDISBASE, flux):
        tmp = dis.coef.clone()
        tmp = dis.coef + flux.coef
        return dis.new(dis.u, dis.theta, tmp)

    def forward(self, dis: HDISBASE, T, dx, Kn, verbose=False):
        if torch.tensor(T).numel() == 1:
            T = torch.tensor(T, device=dis.u.device).repeat(dis.u.shape[0])

        trec = torch.zeros((dis.u.shape[0],), device=dis.u.device)
        while (-trec + T).gt(0).any():
            dt = torch.minimum(self.get_tstep(dis, dx), -trec + T).detach()
            dt = torch.minimum(dt, self.max_dt * torch.ones_like(dt))
            if verbose:
                print(dt, "/", T)
            dis = self.ode_step(dis, dt, dx, Kn)
            trec += dt
        return dis

    def solve_to(self, T, verbose=False):
        dis = self.f0
        dx = self.dx
        if (not torch.is_tensor(T)) and torch.tensor(T).numel() == 1:
            T = torch.tensor(T, device=dis.u.device).repeat(dis.u.shape[0])

        trec = torch.zeros((dis.u.shape[0],), device=dis.u.device)
        while (-trec + T).gt(0).any():
            dt = torch.minimum(self.get_tstep(dis, dx), -trec + T).detach()
            dt = torch.minimum(dt, self.max_dt * torch.ones_like(dt))
            if verbose:
                print(dt, "/", T)
            self.f0 = self.ode_step(dt)
            trec += dt
        return self.f0

    def run(self, dis, T, dx, Kn=None, verbose=False):
        print("running...")
        return self.forward(dis, T, dx, Kn=Kn, verbose=verbose)

    def set_BC_Value(self, LV, RV):
        self.LV = LV  # LV=[rho,u,theta] #shape [[B,1],[B,1],[B,1]]
        self.RV = RV  # RV=[rho,u,theta] #shape [[B,1],[B,1],[B,1]]



'''
Author: lizhengyi lizhengyi@pku.edu.cn
Date: 2022-06-04 20:32:58
LastEditors: lizhengyi lizhengyi@pku.edu.cn
LastEditTime: 2022-06-05 19:26:25
FilePath: /BTE/bte/utils/record.py
Description: save M13 data from DVM
'''
import torch

def collect_data_G13(dis):
    RHO=dis.density().cpu()
    MU=dis.velocity().cpu()
    THETA=dis.temperature().cpu()
    SIGMA=torch.zeros(RHO.shape[:-1]+(3,3))
    for i in range(3):
        for j in range(3):
            idx=[0,0,0]
            idx[i]+=1
            idx[j]+=1
            SIGMA[...,i,j]=dis._M(0,*idx)[...,0]
    Q=torch.zeros(RHO.shape[:-1]+(3,))
    for i in range(3):
        idx=[0,0,0]
        idx[i]+=1
        Q[...,i]=dis._M(1,*idx)[...,0]
    data={
        "rho":RHO,
        "u":MU,
        "theta":THETA,
        "sigma":SIGMA,
        "Q":Q,
    }
    return data

def collect_data_G13Ex(dis,f=None):
    if f is None:
        f=dis.f
    RHO=dis.density().cpu()
    MU=dis.velocity().cpu()
    THETA=dis.temperature().cpu()
    SIGMA=torch.zeros(RHO.shape[:-1]+(3,3))
    for i in range(3):
        for j in range(3):
            idx=[0,0,0]
            idx[i]+=1
            idx[j]+=1
            SIGMA[...,i,j]=dis._M(0,*idx,f)[...,0]
    Q=torch.zeros(RHO.shape[:-1]+(3,))
    for i in range(3):
        idx=[0,0,0]
        idx[i]+=1
        Q[...,i]=dis._M(1,*idx,f)[...,0]
    U0_ijk=torch.zeros(RHO.shape[:-1]+(3,3,3))
    for i in range(3):
        for j in range(3):
            for k in range(3):
                idx=[0,0,0]
                idx[i]+=1
                idx[j]+=1
                idx[k]+=1
                U0_ijk[...,i,j,k]=dis._M(0,*idx,f)[...,0]
    U1_ij=torch.zeros(RHO.shape[:-1]+(3,3))
    for i in range(3):
        for j in range(3):
            idx=[0,0,0]
            idx[i]+=1
            idx[j]+=1
            U1_ij[...,i,j]=dis._M(1,*idx,f)[...,0]
    W2=dis._M(2,0,0,0,f).cpu()
    data={
        "rho":RHO,
        "u":MU,
        "theta":THETA,
        "sigma":SIGMA,
        "Q":Q,
        "U0_ijk":U0_ijk,
        "U1_ij":U1_ij,
        "W2":W2
    }
    return data



import numpy as np

# u_t = f(u) + g(u)
# f(u) non-stiff, explicit
# g(u) stiff, implicit

# Butcher tableau
# c|A
# ---
#  |b

# g(u,args)  BGK part: g=1/tau(M(f)-f)
# f(u,args)  flux and other collision part

# Implicit-Explicit schemes for BGK kinetic equations

# y size [B,C,W]

class IMEX_RK:
    r"""IMEX-RK method

        y_t = f(t,y) + g(t,y)

        # f(t,y) non-stiff, explicit
        # g(t,y) stiff, implicit

        Args:
            b1 ([type]): [description]
            A1 ([type]): [description]
            b2 ([type]): [description]
            A2 ([type]): [description]

        Raises:
            ValueError: A1,b1,A2,b2 shape must be consistent.
    """

    def __init__(self, b1, A1, b2, A2) -> None:
        self.Stage = len(b1)
        if len(b2) != self.Stage:
            raise ValueError("Invalid Butcher tableau")
        if A1.shape[0] != A1.shape[1] or A1.shape[0] != self.Stage:
            raise ValueError("Invalid Butcher tableau")
        if A2.shape[0] != A2.shape[1] or A2.shape[0] != self.Stage:
            raise ValueError("Invalid Butcher tableau")
        self.b1 = b1
        self.A1 = A1
        self.b2 = b2
        self.A2 = A2
        self.c1 = [sum(a1) for a1 in A1]
        self.c2 = [sum(a2) for a2 in A2]

    def __call__(self, t0, y0, h, f, g, g_solver):
        r"""
        __call__ [summary]

        Args:
            f ([type]): [description]
            t0 ([type]): [description]
            y0 ([type]): [description]
            h ([type]): [description]
            g ([type]): [description]
            g_solver ([type]): the solver of u^{(i)}=H(\tilde u^{(i)},\Delta t A_{ii};G), u=g_solver(tilde_u, \Delta t A_{ii})

        Returns:
            [type]: [description]
        """
        f_list = []
        g_list = []
        for s in range(self.Stage):
            res_f = sum([self.A1[s, i] * f_list[i] for i in range(s)])
            res_g = sum([self.A2[s, i] * g_list[i] for i in range(s)])
            tilde_u = y0 + h * (res_f + res_g)
            tnow = t0 + self.c2[s]*h
            y = g_solver(tilde_u, h * self.A2[s, s], tnow)
            f_list.append(f(self.c1[s],y))
            g_list.append(g(self.c2[s],y))
        y = y0 + h * (
            sum([f_list[j] * self.b1[j] for j in range(self.Stage)])
            + sum([(g_list[j]) * self.b2[j] for j in range(self.Stage)])
        )
        return y


class IMEX_RK1(IMEX_RK):
    def __init__(self) -> None:
        b1 = np.array([1])
        A1 = np.array([[0]])
        b2 = np.array([1])
        A2 = np.array([[1]])
        super().__init__(b1, A1, b2, A2)


class IMEX_RK2(IMEX_RK):
    def __init__(self) -> None:
        b1 = np.array([0, 0.5, 0.5])
        A1 = np.array([[0, 0, 0], [0, 0, 0], [0, 1, 0]])
        b2 = np.array([0, 0.5, 0.5])
        A2 = np.array([[0.5, 0, 0], [-0.5, 0.5, 0], [0, 0.5, 0.5]])
        super().__init__(b1, A1, b2, A2)


class IMEX_RK3(IMEX_RK):
    def __init__(self) -> None:
        alpha = 0.24169426078821
        beta = 0.06042356519705
        eta = 0.1291528696059
        b1 = np.array([0, 1 / 6, 1 / 6, 2 / 3])
        A1 = np.array([[0, 0, 0, 0], [0, 0, 0, 0], [0, 1, 0, 0], [0, 1 / 4, 1 / 4, 0]])
        b2 = np.array([0, 1 / 6, 1 / 6, 2 / 3])
        A2 = np.array(
            [
                [alpha, 0, 0, 0],
                [-alpha, alpha, 0, 0],
                [0, 1 - alpha, alpha, 0],
                [beta, eta, 0.5 - alpha - beta - eta, alpha],
            ]
        )
        super().__init__(b1, A1, b2, A2)


#  Stiffly Accurate Scheme, which mean A2[-1,:]=b2
class IMEX_RK1_stiff(IMEX_RK):
    def __init__(self) -> None:
        b1 = np.array([1, 0])
        A1 = np.array([[0, 0], [1, 0]])
        b2 = np.array([0, 1])
        A2 = np.array([[0, 0], [0, 1]])
        super().__init__(b1, A1, b2, A2)


class IMEX_RK2_stiff(IMEX_RK):
    def __init__(self) -> None:
        b1 = np.array([0, 1, 0])
        A1 = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]])
        b2 = np.array([0.5, 0, 0.5])
        A2 = np.array([[0, 0, 0], [0, 1, 0], [0.5, 0.0, 0.5]])
        super().__init__(b1, A1, b2, A2)


import torch

def linear_reconstruction_MinMod(fL:torch.Tensor, f:torch.Tensor, fR:torch.Tensor)->torch.Tensor:
    """linear_reconstruction using MinMod limiter

    Args:
        fL (torch.Tensor): left cell 
        f (torch.Tensor): cell
        fR (torch.Tensor): right cell

    Returns:
        torch.Tensor: leftRec, rightRec
    """
    theta1 = f - fL
    theta2 = fR - f
    slope = (theta1.sign()+theta2.sign())/2*torch.minimum(theta1.abs(),theta2.abs())
    RecL = f - slope / 2 
    RecR = f + slope / 2 
    return RecL, RecR

def linear_reconstruction_VanLeer(fL:torch.Tensor, f:torch.Tensor, fR:torch.Tensor)->torch.Tensor:
    """linear_reconstruction using VanLeer limiter

    Args:
        fL (torch.Tensor): left cell 
        f (torch.Tensor): cell
        fR (torch.Tensor): right cell

    Returns:
        torch.Tensor: leftRec, rightRec
    """
    theta1 = f - fL
    theta2 = fR - f
    slope = (theta1.sign()+theta2.sign())*theta1.abs()*theta2.abs()/(theta1.abs()+theta2.abs()+1e-8)
    RecL = f - slope / 2 
    RecR = f + slope / 2 
    return RecL, RecR


'''
Author: lizhengyi lizhengyi@pku.edu.cn
Date: 2022-06-05 11:05:55
LastEditors: lizhengyi lizhengyi@pku.edu.cn
LastEditTime: 2022-06-05 11:05:56
FilePath: /BTE/bte/utils/gas.py
Description: 这是默认设置,请设置`customMade`, 打开koroFileHeader查看配置 进行设置: https://github.com/OBKoro1/koro1FileHeader/wiki/%E9%85%8D%E7%BD%AE
'''
import math

def get_gamma(ck):
    get_gamma = (ck+5)/(ck+3)
    return get_gamma

def get_potential(omega):
    if(omega==0.5):
        alpha=1.0
    else:
        eta=4.0/(2.0*omega-1.0)+1.0
        alpha=(eta-5.0)/(eta-1.0)
    return alpha

def get_mu(alpha,omega,kn):
    get_mu = 5*(alpha+1)*(alpha+2)*math.sqrt(math.pi)/(4*alpha*(5-2*omega)*(7-2*omega))*kn
    return  get_mu

def get_kn_bzm(alpha,mu_ref):
    kn_bzm=64*math.sqrt(2.0)**alpha/5.0*math.gamma((alpha+3)/2)*math.gamma(2.0)*math.sqrt(math.pi)*mu_ref
    return kn_bzm
    


import torch
import torch.nn.functional as F


def van_leer_limiter(r:torch.Tensor)->torch.Tensor:
    r"""
    van_leer_limiter

    .. math::
        \phi _{{vl}}(r)={\frac  {r+\left|r\right|}{1+\left|r\right|}}
    
    Args:
        r (torch.Tensor): [description]

    Returns:
        torch.Tensor: [description]
    """
    return (r + torch.abs(r)) / (1.0 + torch.abs(r))


def limiter_minmod(theta:torch.Tensor)->torch.Tensor:
    r"""
    limiter_minmod

    Args:
        theta (torch.Tensor): [description]

    .. math::
        \phi _{{mm}}(r)=\max \left[0,\min \left(1,r\right)\right]

    Returns:
        torch.Tensor: [description]
    """
    return F.relu(torch.minimum(torch.ones_like(theta), theta))


def limiter_superbee(theta:torch.Tensor)->torch.Tensor:
    r"""
    limiter_superbee 

    Args:
        theta (torch.Tensor): [description]

    .. math::
        \phi _{{sb}}(r)=\max \left[0,\min \left(2r,1\right),\min \left(r,2\right)\right]

    Returns:
        torch.Tensor: [description]
    """
    return torch.maximum(
        torch.minimum(2 * theta, torch.ones_like(theta)),
        torch.minimum(theta, 2 * torch.ones_like(theta)),
    )


def limiter_mc(theta:torch.Tensor)->torch.Tensor:
    r"""
    limiter_mc 

    Args:
        theta (torch.Tensor): [description]

    .. math::
        \phi _{{mc}}(r)=\max \left[0,\min \left(2r,0.5(1+r),2\right)\right]

    Returns:
        torch.Tensor: [description]
    """
    return torch.maximum(
        torch.zeros_like(theta),
        torch.minimum(
            torch.minimum(2 * theta, 2 * torch.ones_like(theta)), 0.5 * (theta + 1)
        ),
    )


def limiter_zero(theta:torch.Tensor)->torch.Tensor:
    """
    limiter_zero Always return 0. Linear reconstruction with limiter_zero is equal to constant reconstruction.

    .. math::
        \phi _{{0}}(r)=0

    Args:
        theta (torch.Tensor): [description]

    Returns:
        torch.Tensor: zeros tensor shape same with theta.
    """
    return torch.zeros_like(theta)



from functools import partial
from math import factorial

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from bte.dvm import distribution
from bte.dvm.distribution import DVDisMeta_Grid
from bte.dvm.distribution import DVDis, DVDisMeta
from bte.utils import IMEX_RK as IMEX_RK
from bte.utils.limiter import *
from bte.utils.weno5 import *
from bte.dvm.collision import collisioner,collisioner_bgk
from bte.utils.specials import eval_hermitenorm

EPS = 1e-8

def F_pm_2(f, vp, vm, dx, dt, limiter=limiter_minmod):
    df = f[..., 1:, :] - f[..., :-1, :]
    rp = (df[..., :-1, :]) / (df[..., 1:, :] + EPS)
    rm = (df[..., 1:, :]) / (df[..., :-1, :] + EPS)
    phi_p = limiter(rp)
    phi_m = limiter(rm)
    Fm = torch.empty_like(f)
    Fp = torch.empty_like(f)

    Fp[..., 1:-1, :] = 0.5 * phi_p * (1.0) * (df[..., 1:, :])
    Fm[..., 1:-1, :] = 0.5 * phi_m * (-1.0) * (df[..., :-1, :])

    return (
        (df[..., 1:-2, :] + Fp[..., 2:-2, :] - Fp[..., 1:-3, :]) / dx,
        (df[..., 2:-1, :] + Fm[..., 3:-1, :] - Fm[..., 2:-2, :]) / dx,
    )


def F_plm(f, vp, vm, dx, dt, limiter=limiter_minmod):
    df = f[..., 1:, :] - f[..., :-1, :]
    rp = (df[..., :-1, :]) / (df[..., 1:, :] + EPS)
    rm = (df[..., 1:, :]) / (df[..., :-1, :] + EPS)
    phi_p = limiter(rp)
    phi_m = limiter(rm)
    Fm = torch.empty_like(f)
    Fp = torch.empty_like(f)

    Fp[..., 1:-1, :] = 0.5 * phi_p * (1.0 - vp * dt / dx) * (df[..., 1:, :])
    Fm[..., 1:-1, :] = 0.5 * phi_m * (-1.0 - vm * dt / dx) * (df[..., :-1, :])

    return (
        (df[..., 1:-2, :] + Fp[..., 2:-2, :] - Fp[..., 1:-3, :]) / dx,
        (df[..., 2:-1, :] + Fm[..., 3:-1, :] - Fm[..., 2:-2, :]) / dx,
    )

def linear_reconstruction_VanLeer(fL:torch.Tensor, f:torch.Tensor, fR:torch.Tensor)->torch.Tensor:
    """linear_reconstruction using VanLeer limiter

    Args:
        fL (torch.Tensor): left cell 
        f (torch.Tensor): cell
        fR (torch.Tensor): right cell

    Returns:
        torch.Tensor: leftRec, rightRec
    """
    theta1 = f - fL
    theta2 = fR - f
    slope = (theta1.sign()+theta2.sign())*theta1.abs()*theta2.abs()/(theta1.abs()+theta2.abs()+1e-8)
    RecL = f - slope / 2 
    RecR = f + slope / 2 
    return RecL, RecR

def linear_reconstruction_MinMod(fL:torch.Tensor, f:torch.Tensor, fR:torch.Tensor)->torch.Tensor:
    """linear_reconstruction using MinMod limiter

    Args:
        fL (torch.Tensor): left cell 
        f (torch.Tensor): cell
        fR (torch.Tensor): right cell

    Returns:
        torch.Tensor: leftRec, rightRec
    """
    theta1 = f - fL
    theta2 = fR - f
    slope = (theta1.sign()+theta2.sign())/2*torch.minimum(theta1.abs(),theta2.abs())
    RecL = f - slope / 2 
    RecR = f + slope / 2 
    return RecL, RecR

def linear_reconstruction(fL, f, fR, limiter=limiter_minmod):
    theta = torch.zeros_like(f)
    theta1 = f - fL
    theta2 = fR - f
    mask = torch.abs(theta1 * theta2) > 1e-12
    theta[mask] = (theta1 / (theta2 + 1e-12))[mask]

    theta = limiter(theta)
    RecL = f - theta / 2 * theta2
    RecR = f + theta / 2 * theta2

    return RecL, RecR

def get_reflecting_bdv(dis:DVDis, BV, xi, side, no_linear=False):
    """[summary]

    Args:
        dis (DVDis): [description]
        BV ([type]): [description]
        xi ([type]): [description]
        side ([type]):side=-1 for left BV, side=1 for right BV
        no_linear (bool, optional): [description]. Defaults to False.

    Raises:
        ValueError: [description]

    Returns:
        [type]: [description]
    """
    if side == -1:
        fl = dis.f
        fbl_toleft = dis.one_side(fl, -1)
        fbl_toright = dis.reverse(fbl_toleft)
        fbl_maxwellian = dis.maxwellian_half(rho_u_theta=BV, sign=1)
        # fbl_maxwellian的动量应该和fbl_toleft相等。
        fbl_maxwellian_U = dis.rho_u_theta(fbl_maxwellian)
        fbl_toleft_U = dis.rho_u_theta(fbl_toleft)
        c = (
            (
                (fbl_toleft_U[0] * fbl_toleft_U[1])
                / (fbl_maxwellian_U[0] * fbl_maxwellian_U[1])
            )
            .abs()
        )
        fbl = (
            xi * c * fbl_maxwellian + (1 - xi) * fbl_toright + fbl_toleft
        )
        if no_linear:
            fbll=fbl
        else:
            fbll=2*fbl-fl
        return DVDis(dis.v_meta,fbll)
    elif side == 1:
        fr = dis.f
        fbr_toright = dis.one_side(fr, 1)
        fbr_toleft = dis.reverse(fbr_toright)
        fbr_maxwellian = dis.maxwellian_half(rho_u_theta=BV, sign=-1)
        # fbl_maxwellian的动量应该和fbl_toleft相等。
        fbr_maxwellian_U = dis.rho_u_theta(fbr_maxwellian)
        fbr_toright_U = dis.rho_u_theta(fbr_toright)
        c = (
            (
                (fbr_toright_U[0] * fbr_toright_U[1])
                / (fbr_maxwellian_U[0] * fbr_maxwellian_U[1])
            )
            .abs()
        )
        fbr = (
            xi * c * fbr_maxwellian + (1 - xi) * fbr_toleft + fbr_toright
        )
        if no_linear:
            fbrr=fbr
        else:
            fbrr=2*fbr-fr
        return DVDis(dis.v_meta,fbrr)
    else:
        raise ValueError


class Grid_onDVDis_NU(DVDis):
    """A grid class that stores the details and solution of the
    computational grid."""

    def __init__(self, xmin, xmax, nx, vmin, vmax, nv, v_discrete="uni", device="cpu"):
        self.xmin = xmin
        self.xmax = xmax
        self.vmin = vmin
        self.vmax = vmax
        self.nx, self.nv = nx, nv
        dx = float(self.xmax - self.xmin) / self.nx
        self.v_discrete=v_discrete
        x = torch.arange(
            start = self.xmin + dx / 2, end = self.xmax + dx / 2, step=dx
        )
        #self.dx = dx * torch.ones(nx,1)
        # self.dx = self.dx.to(device=device)
        # self.x = x.to(device=device)


        if isinstance(nv, int):# If single dimension
            if v_discrete == "leg":
                v, w = distribution.legendre_velocity(nv, vmin, vmax)
            elif v_discrete == "uni":
                v, w = distribution.uniform_velocity(nv, vmin, vmax)
            else:
                raise ValueError
            # v = v.to(device=device)
            # w = w.to(device=device)
            v_meta=DVDisMeta(v, w)
        else:# If multiple dimension
            if v_discrete == "leg":
                vL, wL = distribution.velocity_list(distribution.legendre_velocity,nv, vmin, vmax)
            elif v_discrete == "uni":
                vL, wL = distribution.velocity_list(distribution.uniform_velocity,nv, vmin, vmax)
            else:
                raise ValueError
        v_meta=DVDisMeta_Grid(vL, wL)
        super().__init__(v_meta, None)

        self.register_buffer("x",x)
        self.register_buffer("dx",dx * torch.ones(nx,1))

    # def cuda(self):
    #     self.v_meta=self.v_meta.cuda()
    #     self.dx=self.dx.cuda()
    #     self.x=self.x.cuda()
    #     return self



# BGKSolver with MUSL flux (一种DVM方法)
class BGKSolver(nn.Module):
    def __init__(
        self,
        xmin,
        xmax,
        nx,
        vmin,
        vmax,
        nv,
        v_discrete="uni",
        BC_type="constant",
        bgk_simple_kn="simple",
        device="cpu",
    ):
        super().__init__()
        self.kn = np.nan

        self.set_BC(BC_type)
        self.BC = BC_type
        self.set_space_order(2)
        self.set_time_stepper("bgk-RK2")
        self.bgk_simple_kn = bgk_simple_kn
        self.dis = Grid_onDVDis_NU(
            xmin, xmax, nx, vmin, vmax, nv, v_discrete=v_discrete, device=device
        )

        abs_v = torch.abs(self.dis.v)
        self.register_buffer("_v_p", 0.5 * (self.dis.v + abs_v))
        self.register_buffer("_v_m", 0.5 * (self.dis.v - abs_v))


    @property
    def f(self):
        return self.dis.f

    @f.setter
    def f(self, value):
        self.dis.f = value

    def set_order(self, space_order, time_order, time_stiff=False):
        if space_order == "plm" and (time_order != 1 or time_stiff):
            raise ValueError("plm only suit with RK1")
        self.set_space_order(space_order)
        self.set_time_order(time_order, time_stiff)

    def set_space_order(self, order):
        if order == 1:
            self._grad_vf = self._grad_vf_constant
        elif order == 2:
            self._grad_vf = self._grad_vf_2
        elif order == 5:
            self._grad_vf = self._grad_vf_weno
        elif order == "plm":
            self._grad_vf = self._grad_vf_plm
        else:
            raise ValueError

    def set_time_stepper(self,method:"str"):
        """set time stepper

        Args:
            method (str): can be "bgk-RK1","bgk-RK2","bgk-RK3","bgk-RKs1","bgk-RKs2"
            "Euler", "IMEX-1", "IMEX-2", "IMEX"
        """
        bgk_RK={
            "bgk-RK1":IMEX_RK.IMEX_RK1(),
            "bgk-RK2":IMEX_RK.IMEX_RK2(),
            "bgk-RK3":IMEX_RK.IMEX_RK3(),
            "bgk-RK1s":IMEX_RK.IMEX_RK1_stiff(),
            "bgk-RK2s":IMEX_RK.IMEX_RK2_stiff()
        }
        if method in ["bgk-RK1","bgk-RK2","bgk-RK3","bgk-RK1s","bgk-RK2s"]:
            self.step=self.step_rk
            self.RK=bgk_RK[method]
        elif method=="Euler":
            self.step=self.step_euler
        elif method=="IMEX":
            self.step=self.step_imex
        elif method=="IMEX-1":
            self.step=self.step_imex1
        else:
            raise ValueError

    def set_collisioner(self,collision_type:str,**args):
        """set collision type

        Args:
            collision_type (str): "BGK" or "FFT"

        Raises:
            ValueError: _description_
        """
        if collision_type=="FFT":
            self.collision_type="FFT"
            self.coller=collisioner(self.dis.v_meta, quad_num=64, **args)
        elif collision_type=="BGK":
            self.collision_type="BGK"
            self.coller=collisioner_bgk(self.dis.v_meta,**args)
        else:
            raise ValueError

    def set_initial(self, kn, f0):
        if isinstance(kn, torch.Tensor):
            self.kn = kn
        else:
            self.kn = kn * torch.ones(f0.shape[:-1]+(1,),device=f0.device).expand_as(f0)
        self.dis.f = f0.clone()
        self.left_BC = self.dis.f[..., 0:1, :].clone()
        self.right_BC = self.dis.f[..., -1:, :].clone()

    def set_BC(self, BC):
        self.BC = BC
        if BC == "periodic":
            self.apply_BC = self.padding_periodic
        elif BC == "cauchy":
            self.apply_BC = self.padding_replicate
        elif BC == "reflecting":
            #self.apply_BC = self.padding_reflecting_linear
            self.apply_BC = self.padding_replicate
            #self.xi = 1.0
        elif BC == "constant":
            self.apply_BC = self.padding_constant
        elif BC == "evaporation":
            self.apply_BC = self.padding_evaporation
        elif BC == "evaporation2":
            self.apply_BC = self.padding_evaporation2
        else:
            raise NotImplementedError

    def set_BC_Value(self, LV, RV):
        self.LV = LV  # LV=[rho,u,theta] #shape [[B,1,1],[B,1,1],[B,1,1]]
        self.RV = RV  # RV=[rho,u,theta] #shape [[B,1,1],[B,1,1],[B,1,1]]

    def padding_periodic(self, f):
        f = f.transpose(-1, -2)
        return F.pad(
            f, pad=(self.padding_order, self.padding_order), mode="circular"
        ).transpose(-1, -2)

    def padding_replicate(self, f):
        f = f.transpose(-1, -2)
        return F.pad(
            f, pad=(self.padding_order, self.padding_order), mode="replicate"
        ).transpose(-1, -2)

    def padding_constant(self, f):
        f = f.transpose(-1, -2)
        fbc = F.pad(
            f, pad=(self.padding_order, self.padding_order), mode="constant"
        ).transpose(-1, -2)
        fbc[..., 0 : self.padding_order, :] = self.left_BC
        fbc[..., -self.padding_order :, :] = self.right_BC
        return fbc

    def get_reflecting_bdv(self, fb, side):
        return get_reflecting_bdv(DVDis(self.dis.v_meta,fb),
                                (self.LV if side<0 else self.RV),
                                self.xi,side,no_linear=True).f
        #print("falg")
        # side=-1 for left boundary, side=+1 for right boundary
        # if side == -1:
        #     fl = fb
        #     fbl_toleft = self.dis.one_side(fl, -1)
        #     fbl_toright = self.dis.reverse(fbl_toleft)
        #     fbl_maxwellian = self.dis.maxwellian_half(rho_u_theta=self.LV, sign=1)
        #     # fbl_maxwellian的动量应该和fbl_toleft相等。
        #     # print(fbl_maxwellian.shape,fbl_toleft.shape)
        #     fbl_maxwellian_U = self.dis.rho_u_theta(fbl_maxwellian)
        #     fbl_toleft_U = self.dis.rho_u_theta(fbl_toleft)
        #     c = (
        #         (
        #             (fbl_toleft_U[0] * fbl_toleft_U[1])
        #             / (fbl_maxwellian_U[0] * fbl_maxwellian_U[1])
        #         )
        #         .abs()
        #     )
        #     fbl = (
        #         self.xi * c * fbl_maxwellian + (1 - self.xi) * fbl_toright + fbl_toleft
        #     )
        #     return fbl
        # elif side == 1:
        #     fr = fb
        #     fbr_toright = self.dis.one_side(fr, 1)
        #     fbr_toleft = self.dis.reverse(fbr_toright)
        #     fbr_maxwellian = self.dis.maxwellian_half(rho_u_theta=self.RV, sign=-1)
        #     # fbl_maxwellian的动量应该和fbl_toleft相等。
        #     fbr_maxwellian_U = self.dis.rho_u_theta(fbr_maxwellian)
        #     fbr_toright_U = self.dis.rho_u_theta(fbr_toright)
        #     c = (
        #         (
        #             (fbr_toright_U[0] * fbr_toright_U[1])
        #             / (fbr_maxwellian_U[0] * fbr_maxwellian_U[1])
        #         )
        #         .abs()
        #     )
        #     fbr = (
        #         self.xi * c * fbr_maxwellian + (1 - self.xi) * fbr_toleft + fbr_toright
        #     )
        #     return fbr
        # else:
        #     raise ValueError

    def getTau(self,U):
        if self.bgk_simple_kn=="simple":
            tau= self.kn 
        elif self.bgk_simple_kn=="simple2":
            tau= self.kn / U[0] 
        elif self.bgk_simple_kn=="mu_ref":
            tau=  self.mu_ref*2/(U[2]*2)**(1-self.omega)/U[0]
        else:
            raise ValueError
        return tau

    def update(self, qP, qN):
        if self.BC == "reflecting":
            fl = qN[..., 1:2, :]
            fbl = self.get_reflecting_bdv(fl, -1)
            qP[..., 0:1, :] = fbl

            fr = qP[..., -2:-1, :]
            fbr = self.get_reflecting_bdv(fr, 1)
            qN[..., -1:, :] = fbr
        return qP, qN

    def padding_reflecting_linear(self, f):
        if self._grad_vf == self._grad_vf_constant:
            return self.padding_replicate(f)
        # f in shape [B,W,V]
        fbl = f[:, 0:1, :]
        fblr = f[:, 1:2, :]
        fbr = f[:, -1:, :]
        fbrl = f[:, -2:-1, :]

        fblp = [fbl + (i + 1) * (fbl - fblr) for i in range(self.padding_order)]
        fbrp = [fbr + (i + 1) * (fbr - fbrl) for i in range(self.padding_order)]

        fbc = torch.cat(fblp[::-1] + [f,] + fbrp, dim=-2)

        return fbc

    def padding_reflecting(self, f):
        # f in shape [B,W,V]
        fbl = f[:, 0:1, :]
        fbr = f[:, -1:, :]

        # calculate fbl
        fl = f[:, 0:1, :]
        fbl = self.get_reflecting_bdv(fl, -1)

        # calculate fbl
        fr = f[:, -1:, :]
        fbr = self.get_reflecting_bdv(fr, 1)

        fbc = torch.cat(
            [fbl,] * self.padding_order + [f,] + [fbr,] * self.padding_order, dim=1
        )
        return fbc

    def padding_evaporation(self,f):
        fbl = f[:, 0:1, :]
        fblr = f[:, 1:2, :]
        fbr = f[:, -1:, :]
        fbrl = f[:, -2:-1, :]

        fblp_gas = [fbl + (i + 1) * (fbl - fblr) for i in range(self.padding_order)]
        fbrp_gas = [fbr + (i + 1) * (fbr - fbrl) for i in range(self.padding_order)]

        fbl_vapor=self.dis.maxwellian(rho_u_theta=self.LV)
        fbr_vapor=self.dis.maxwellian(rho_u_theta=self.RV)

        fblp=[g*(self.dis.v[...,0]<0)+fbl_vapor*(self.dis.v[...,0]>=0) for g in fblp_gas]
        fbrp=[g*(self.dis.v[...,0]>0)+fbr_vapor*(self.dis.v[...,0]<=0) for g in fbrp_gas]

        fbc = torch.cat(fblp[::-1] + [f,] + fbrp, dim=1)
        return fbc

    def padding_evaporation2(self,f):
        fbl = f[:, 0:1, :]
        fblr = f[:, 1:2, :]
        fbr = f[:, -1:, :]
        fbrl = f[:, -2:-1, :]

        fblp_gas = [fbl + (i + 1) * (fbl - fblr) for i in range(self.padding_order)]
        fbrp_gas = [fbr + (i + 1) * (fbr - fbrl) for i in range(self.padding_order)]

        fbl_vapor=self.dis.maxwellian(rho_u_theta=self.LV)
        fbr_vapor=self.dis.maxwellian(rho_u_theta=self.RV)

        fblp=[g*(self.dis.v[...,0]<0)+fbl_vapor*(self.dis.v[...,0]>0) for g in fblp_gas]
        fbrp=[g*(self.dis.v[...,0]>0)+fbr_vapor*(self.dis.v[...,0]<0) for g in fbrp_gas]

        fbc = torch.cat(fblp + [f,] + fbrp, dim=1)
        return fbc

    def upwind(self, fl, fr, direction=0):
        return self._v_p[..., direction] * fl + self._v_m[..., direction] * fr

    def _grad_vf_constant(self, f, dt):
        self.padding_order = 1
        fbc = self.apply_BC(f)
        qP = fbc  # j+1/2
        qN = fbc  # j-1/2
        qP, qN = self.update(qP, qN)
        fluxL = self.upwind(qP[..., 0:-2, :], qN[..., 1:-1, :])
        fluxR = self.upwind(qP[..., 1:-1, :], qN[..., 2:, :])
        return 1 / self.dis.dx * (fluxR - fluxL)

    def _grad_vf_2(self, f, dt):
        self.padding_order = 2
        fbc = self.apply_BC(f)
        qN, qP = linear_reconstruction_VanLeer(
            fbc[..., 0:-2, :], fbc[..., 1:-1, :], fbc[..., 2:, :]
        )
        qP, qN = self.update(qP, qN)
        fluxL = self.upwind(qP[..., 0:-2, :], qN[..., 1:-1, :])
        fluxR = self.upwind(qP[..., 1:-1, :], qN[..., 2:, :])
        return 1 / self.dis.dx * (fluxR - fluxL)

    def _grad_vf_weno(self, f, dt):
        self.padding_order = 3
        fbc = self.apply_BC(f)
        qN, qP = weno5_LR(
            fbc[..., :-4, :],
            fbc[..., 1:-3, :],
            fbc[..., 2:-2, :],
            fbc[..., 3:-1, :],
            fbc[..., 4:, :],
        )
        qP, qN = self.update(qP, qN)
        fluxL = self.upwind(qP[..., 0:-2, :], qN[..., 1:-1, :])
        fluxR = self.upwind(qP[..., 1:-1, :], qN[..., 2:, :])
        return 1 / self.dis.dx * (fluxR - fluxL)

    def _grad_vf_plm(self, f, dt):
        self.padding_order = 2
        fbc = self.apply_BC(f)
        Fp, Fm = F_plm(fbc, self._v_p, self._v_m, self.dis.dx, dt)
        return self._v_p * Fp + self._v_m * Fm

    # def half_maxwellian(self, arg, sign):
    #     # 生成以arg为参数的一个maxwellian，并截断保留一半
    #     rho, u, T = arg
    #     maxwellian = (rho / torch.sqrt(2 * pi * T))[..., None] * torch.exp(
    #         -((u[..., None] - self.v) ** 2) / (2 * T)[..., None]
    #     )
    #     # B,W,V
    #     if sign > 0:
    #         maxwellian[..., self.dis.v[...,0] < 0] = 0
    #     elif sign < 0:
    #         maxwellian[..., self.dis.v[...,0] > 0] = 0
    #     else:
    #         raise ValueError("sign should be 1 or -1")
    #     return maxwellian

    # def cut_distribution(self, f, sign):
    #     f = f.clone()
    #     if sign == 1:
    #         f[..., self.v < 0] = 0
    #     elif sign == -1:
    #         f[..., self.v > 0] = 0
    #     else:
    #         raise ValueError("sign should be 1 or -1")
    #     return f

    # def reverse_distribution(self, f):
    #     return torch.flip(f, [-1,])


    def step_rk(self, dt):
        f = self.f
        Flux = partial(self._grad_vf, dt=dt)
        ff = lambda x: -Flux(x)
        getTau=lambda U: self.getTau(U)
        f = self.RK(ff, f, dt, getTau, self.dis.rho_u_theta, self.dis.maxwellian)
        self.f = f

    def step_imex(self, dt):
        f = self.f
        flux = self._grad_vf(f,dt)
        ftmp = f - dt*flux
        Unew=self.dis.rho_u_theta(ftmp)
        Tau=self.getTau(Unew)
        Mnew=self.dis.maxwellian(Unew)
        #print("Unew",Unew[0][0,:,0],Unew[1][0,:,0],Unew[2][0,:,0])
        expt=torch.exp(-dt/Tau)
        #print(expt.flatten()[0],dt,Tau.flatten()[0])
        Q=self.coller.do_collision(f,self.kn)
        QU=self.dis.rho_u_theta(Q)
        #print("QU",QU[0][0,:,0],QU[1][0,:,0],QU[2][0,:,0])
        fnew=(ftmp+dt*((1-expt)/Tau*Mnew+Q*expt))/(1+dt/Tau*(1-expt))
        self.f=fnew
        return fnew

    def solve_to(self, tmax, max_dt=None):
        t = 0
        if max_dt is None:
            max_dt = 0.45 *  self.dis.dx / self.dis.vmax[0]
        while t < tmax:
            dt = min(tmax - t, max_dt)
            self.step(dt)  # 使用算法迭代一步
            t += dt

    def step_euler(self, dt):
        f = self.f
        flux = self._grad_vf(f,dt)
        ftmp = f - dt*flux
        Unew = self.dis.rho_u_theta(ftmp)
        # if self.bgk_simple_kn:
        #     Tau = self.kn
        # else:
        #     Tau = self.kn / Unew[0] 
        Tau=self.getTau(Unew)
        Q = self.coller.do_collision(f, Tau)
        self.Q=Q
        fnew=ftmp+dt*Q
        self.f=fnew
        return fnew

    def get_gas_macro(self):
        """返回(rho, m, E)

        Returns:
            torch.array: (rho, m, E)
        """
        rho = self.density()
        u = self.velocity()
        T = self.temperature()
        m = rho * u
        E = 0.5 * (u ** 2 + T) * rho
        return torch.vstack((rho, m, E))

    def get_hermite_moments(self, M_order):
        u, sqr_T = self.velocity(), torch.sqrt(self.temperature())
        c_v = (self.v[None, None, :] - u[:, :, None]) / sqr_T[:, :, None]
        mom = torch.zeros(u.shape + (M_order + 2,))
        mom[..., 0] = u
        mom[..., 1] = sqr_T ** 2
        for i in range(M_order):
            mom[..., i + 2] = (
                self.sum(self.f * eval_hermitenorm(i, c_v)) / factorial(i) * sqr_T ** i
            )
        return mom

    def get_f(self):
        return self.f.clone()

    def get_entropy(self):
        return self.sum(self.f * torch.log(self.f))


from __future__ import annotations
from typing import Callable, Tuple, List
import torch
import torch.nn.functional as F
import torch.nn as nn
from math import factorial, gamma, sqrt
import math
import numpy as np
import abc
from bte.utils.indexs import *
from bte.utils.operators import make_tensor

class distributionBase(metaclass=abc.ABCMeta):
    """
    The base class for gas distribution.
    """
    def __init__(self):
        pass

    @abc.abstractmethod
    def density(self):
        pass

    @abc.abstractmethod
    def velocity(self):
        pass

    @abc.abstractmethod
    def temperature(self):
        pass

    @abc.abstractmethod
    def __add__(self, anotherDis):
        pass

    @abc.abstractmethod
    def __radd__(self, anotherDis):
        pass

    @abc.abstractmethod
    def __sub__(self, anotherDis):
        pass

    @abc.abstractmethod
    def __mul__(self, multiplier):
        pass

    @abc.abstractmethod
    def __rmul__(self, multiplier):
        pass

    @abc.abstractmethod
    def __truediv__(self, divider):
        pass


class HermitedistributionBase():
    def __init__(self):
        pass

    @abc.abstractmethod
    def down(self):
        pass

    @abc.abstractmethod
    def new(self, *args, **kw):
        pass


class HermiteDis(HermitedistributionBase):
    """
    HermiteDis: Gas distribution in Hermite form.
    """

    def __init__(self, u:torch.Tensor, theta:torch.Tensor, coef:torch.Tensor):
        self.u = u
        self.theta = theta
        self.coef = coef
        
    def new(self, u, theta, coef, *args, **kw):
        return HermiteDis(u, theta, coef)

    def __str__(self) -> str:
        return "u:"+self.u.__str__()+"\n"+"theta:"+self.theta.__str__()+"\n"+"coef:"+self.coef.__str__()

    @staticmethod
    def empty(order: int,device='cpu') -> "HermiteDis":
        u = torch.empty((0, 1),device=device)
        theta = torch.empty((0, 1),device=device)
        coef = torch.empty((0, order),device=device)
        return HermiteDis(u, theta, coef)

    def from_DVDis(self, dis: "DVDis", u=None,theta=None) -> "HermiteDis":
        if u is None:
            u = dis.velocity()
        if theta is None:
            theta = dis.temperature()
        sqr_T = torch.sqrt(theta)
        lenv = dis.v.shape[0]
        # print(u.shape,theta.shape,dis.v.shape,u.shape[:-1]+(lenv,))

        c_v = (dis.v[...,0] - u.expand(u.shape[:-1] + (lenv,))) / sqr_T.expand(
            sqr_T.shape[:-1] + (lenv,)
        )
        # print(u.shape[:-1] + (lenv,))
        # print(c_v.shape)
        M = self.ORDER
        mom = torch.empty(dis.f.shape[:-1] + (M,),device=u.device)
        for i in range(M):
            #print(dis.f.shape, eval_hermitenorm(i, c_v).shape,sqr_T.shape)
            #print(dis.sum(dis.f * eval_hermitenorm(i, c_v)).shape)
            tmp = (
                dis.sum(dis.f * eval_hermitenorm(i, c_v))
                / factorial(i)
                * sqr_T ** i
            )
            #print(tmp.shape)
            mom[..., i : i + 1] = tmp
        return HermiteDis(u, theta, mom)

    def cpu(self):
        return HermiteDis(self.u.cpu(), self.theta.cpu(), self.coef.cpu())

    def cuda(self):
        return HermiteDis(self.u.cuda(), self.theta.cuda(), self.coef.cuda())

    def to(self, args):
        return HermiteDis(self.u.to(args), self.theta.to(args), self.coef.to(args))

    @property
    def ORDER(self):
        return self.coef.shape[-1]

    @property
    def rho(self):
        return self.coef[..., 0:1]

    def to_tensor(self):
        return torch.cat([self.u, self.theta, self.coef], dim=-1)

    @property
    def moments(self):
        u, theta, coef = self.u, self.theta, self.coef
        c = u
        s2 = theta
        v0 = coef[..., 0:1]
        v1 = c * v0 + coef[..., 1:2]
        tmp1 = c * v1
        tmp2 = c * c
        tmp3 = coef[..., 2:3]
        v2 = tmp1 + tmp3 + 0.5 * (1.0 * s2 - tmp2) * v0
        return v0, v1, v2

    def density(self):
        return self.coef[...,0:1]

    def velocity(self):
        return self.u

    def temperature(self):
        return self.theta

    @property
    def rho_u_theta(self):
        m0, m1, m2 = self.moments
        v0 = m0
        v1 = m1 / m0
        v2 = 2.0 * m2 / v0 - v1 * v1
        return v0, v1, v2

    def as_tuple(self):
        return self.u, self.theta, self.coef

    def __add__(self, anotherDis):
        return HermiteDis(self.u, self.theta, self.coef + anotherDis.coef)

    def __radd__(self, anotherDis):
        return HermiteDis(self.u, self.theta, self.coef + anotherDis.coef)

    def __sub__(self, anotherDis):
        return HermiteDis(self.u, self.theta, self.coef - anotherDis.coef)

    def __mul__(self, multiplier):
        return HermiteDis(self.u, self.theta, multiplier * self.coef)

    def __rmul__(self, multiplier):
        return HermiteDis(self.u, self.theta, multiplier * self.coef)

    def __truediv__(self, divider):
        return HermiteDis(self.u, self.theta, self.coef / divider)

    def roll(self, idx):
        return HermiteDis(
            d3roll(self.u, idx), d3roll(self.theta, idx), d3roll(self.coef, idx)
        )

    def clone(self):
        return HermiteDis(self.u.clone(), self.theta.clone(), self.coef.clone())

    def mulvec(self):
        tmp = torch.zeros_like(self.coef)
        coef = self.coef
        ind = torch.arange(coef.shape[2] - 1, device=coef.device) + 1
        ind = ind.reshape([1, 1, coef.shape[2] - 1])
        tmp[..., 0:] = tmp[..., 0:] + coef[..., 0:] * self.u
        tmp[..., 1:] = tmp[..., 1:] + coef[..., 0:-1] * self.theta
        tmp[..., 0:-1] = tmp[..., 0:-1] + coef[..., 1:] * ind
        return HermiteDis(self.u, self.theta, tmp)

    def down(self):
        return HermiteDis(self.u, self.theta, self.coef[..., :-1])

    def Project_to_STD(self, STEP: int = 1):
        """
        Project_to_STD 将分布参数重设，重新投影到最佳参数上
        """
        _, u, theta = self.rho_u_theta
        dis = Project_RK4(self, u, theta, STEP=STEP)
        return dis

    def Project_2B(self, dest: "HermiteDis", STEP: int = 1) -> "HermiteDis":
        """
        Project_A2B 将分布投影到dest对应的参数上
        """
        u, theta = dest.u, dest.theta
        dis = Project_RK4(self, u, theta, STEP=STEP)
        return dis


def d3roll(tensor, idx):
    return torch.cat((tensor[..., idx:, :], tensor[..., :idx, :]), -2)


def Project_RK4_helper(dis, u0, theta0, u, theta):
    tmp = (u0 - u) * F.pad(dis[..., :-1], (1, 0)) + 0.5 * (theta0 - theta) * F.pad(
        dis[..., :-2], (2, 0)
    )
    return tmp


def Project_RK4_step(dis, u0, theta0, u, theta, h: float):
    k1 = Project_RK4_helper(dis, u0, theta0, u, theta)
    k2 = Project_RK4_helper(dis + h / 2 * k1, u0, theta0, u, theta)
    k3 = Project_RK4_helper(dis + h / 2 * k2, u0, theta0, u, theta)
    k4 = Project_RK4_helper(dis + h * k3, u0, theta0, u, theta)
    dis = dis + (k1 + 2.0 * k2 + 2.0 * k3 + k4) * h / 6.0
    return dis


def Project_RK4(dis: HermiteDis, u, theta, STEP: int = 1):
    """Project_RK4 使用R-K4格式对分布函数进行投影

    Parameters
    ----------
    dis : torch.Tensor
        分布函数
    u : torch.Tensor
        分布中心
    theta : torch.Tensor
        分布宽度
    STEP : int, optional
        RK方法的步数, by default 1
    Returns
    -------
    torch.Tensor
        投影后的分布函数
    """
    h = 1.0 / STEP
    tmpslice = dis.coef
    u0 = dis.u
    theta0 = dis.theta
    for i in range(STEP):
        tmpslice = Project_RK4_step(tmpslice, u0, theta0, u, theta, h)
    return HermiteDis(u, theta, tmpslice)


def Project_RK4_helperT(dis, u0, theta0, u, theta):
    tmp = (u0 - u) * F.pad(dis[..., 1:], (0, 1)) + 0.5 * (theta0 - theta) * F.pad(
        dis[..., 2:], (0, 2)
    )
    return tmp


def Project_RK4_stepT(dis, u0, theta0, u, theta, h: float = 1):
    k1 = Project_RK4_helperT(dis, u0, theta0, u, theta)
    k2 = Project_RK4_helperT(dis + h / 2 * k1, u0, theta0, u, theta)
    k3 = Project_RK4_helperT(dis + h / 2 * k2, u0, theta0, u, theta)
    k4 = Project_RK4_helperT(dis + h * k3, u0, theta0, u, theta)
    dis = dis + (k1 + 2.0 * k2 + 2.0 * k3 + k4) * h / 6.0
    return dis


def Project_RK4T(dis, u0, theta0, u, theta, STEP: int = 1):
    h = 1.0 / STEP
    tmpslice = dis
    for i in range(STEP):
        tmpslice = Project_RK4_stepT(tmpslice, u0, theta0, u, theta, h)
    return tmpslice


def eval_hermitenorm(n: int, x: torch.Tensor):
    if n < 0:
        return float("nan") * torch.ones_like(x)
    elif n == 0:
        return torch.ones_like(x)
    elif n == 1:
        return x
    else:
        y3 = torch.zeros_like(x)
        y2 = torch.ones_like(x)
        for k in range(n, 1, -1):
            y1 = x * y2 - k * y3
            y3 = y2
            y2 = y1
        return x * y2 - y3

def eval_hermitenorm_seq(n: int, x: torch.Tensor):
    y=torch.zeros((n+1,)+x.shape)
    if n < 0:
        raise ValueError
    if n >= 0:
        y[0]=torch.ones_like(x)
    if n >= 1:
        y[1]=x
    if n >= 2:
        for j in range(2,n+1):
            y[j]=x*y[j-1]-(j-1)*y[j-2]
    return y

class DVDisMeta(nn.Module):
    def __init__(self, v, v_w, *args):
        super().__init__()
        self.register_buffer("v", v)
        self.register_buffer("v2", v ** 2)
        self.register_buffer("v3", v ** 3)
        self.register_buffer("w", v_w)

class DVDisMeta_Grid(DVDisMeta):
    def __init__(self, vL:List, v_wL:List, *args):
        nn.Module.__init__(self)
        assert len(vL)==len(v_wL)
        self.v_dims = nn.ModuleList([DVDisMeta(v,v_w) for v,v_w in zip(vL,v_wL)])
        self.dim = len(vL)
        V, W = product_meshgrid(vL,v_wL)
        self.register_buffer("v", V)
        self.register_buffer("v2", V ** 2)
        self.register_buffer("v3", V ** 3)
        self.register_buffer("w", W)


class DVDis(distributionBase):
    """
    The gas distribution for discrete velocity method.
    """

    def __init__(self, v_meta:DVDisMeta, data):
        super().__init__()
        self.v_meta = v_meta
        self.ndim = v_meta.v.shape[-1]
        self.f = data

    @property
    def v(self):
        return self.v_meta.v

    @property
    def v2(self):
        return self.v_meta.v2

    @property
    def v3(self):
        return self.v_meta.v3

    @property
    def w(self):
        return self.v_meta.w

    def cpu(self):
        return DVDis(self.v_meta.cpu(), self.f.cpu())

    def cuda(self):
        return DVDis(self.v_meta.cuda(), self.f.cuda())

    def to(self, args):
        return DVDis(self.v_meta.to(args), self.f.to(args))

    def __add__(self, anotherDis:DVDis):
        return DVDis(self.v_meta,self.f+anotherDis.f)

    def __radd__(self, anotherDis:DVDis):
        return DVDis(self.v_meta,self.f+anotherDis.f)

    def __sub__(self, anotherDis):
        return DVDis(self.v_meta,self.f-anotherDis.f)

    def __mul__(self, multiplier):
        return DVDis(self.v_meta,self.f*multiplier)

    def __rmul__(self, multiplier):
        return DVDis(self.v_meta,self.f*multiplier)

    def __truediv__(self, divider):
        return DVDis(self.v_meta,self.f/ divider)

    @staticmethod
    def empty(nv: int, vmin: float, vmax: float, method="uni",device='cpu') -> DVDis:
        """An empty new distribution. with velocity grid created by method 'uni' or 'leg'.

        Args:
            nv (int): Numbers of grid
            vmin (float): vmin
            vmax (float): vmax
            method (str, optional): 'uni' for uniform_velocity, 'leg' for legendre_velocity. Defaults to 'uni'.

        Raises:
            ValueError: [description]

        Returns:
            DVDis: [description]
        """
        if method == "uni":
            v, w = uniform_velocity(nv, vmin, vmax)
        elif method == "leg":
            v, w = legendre_velocity(nv, vmin, vmax)
        else:
            raise ValueError(f"method must be uni pof leg, but is {method}.")
        v=v.to(device=device)
        w=w.to(device=device)
        meta = DVDisMeta(v, w)
        return DVDis(meta, None)

    def sum(self, f: torch.Tensor):
        return f @ self.w  # .sum(dim=(-1))
        # return torch.einsum("...ij,...ij->...", f, self.w)

    def density(self, f=None):
        """Compute the macro quantity: density
        """
        if f is None:
            f = self.f
        density = self._m0(f)
        return density

    def velocity(self, f=None):
        """Compute the macro quantity: velocity
        """
        if f is None:
            f = self.f
        velocity = self._m1(f) / self._m0(f)
        return velocity

    def temperature(self, f=None):
        """Compute the macro quantity: temperature
        """
        if f is None:
            f = self.f
        m0 = self._m0(f)
        m1 = self._m1(f)
        v = m1 / m0
        v2 = (v ** 2).sum(dim=-1, keepdim=True)
        temperature = (self._m2(f).sum(dim=-1, keepdim=True) / m0 - v2) / self.v.shape[
            -1
        ]
        return temperature

    def _m0(self, f=None):
        r"""0-th order momentum of f"""
        if f is None:
            f = self.f
        return self.sum(f)

    def _m1(self, f=None):
        r"""1st order momentum of f"""
        if f is None:
            f = self.f
        return torch.cat(
            [self.sum(f * self.v[..., i]) for i in range(self.v.shape[-1])], dim=-1
        )

    def _m2(self, f=None):
        r"""2nd order momentum of f"""
        if f is None:
            f = self.f
        return torch.cat(
            [self.sum(f * self.v2[..., i]) for i in range(self.v.shape[-1])], dim=-1
        )

    def _m3(self, f=None):
        r"""3nd order momentum of f"""
        if f is None:
            f = self.f
        return torch.cat(
            [self.sum(f * self.v3[..., i]) for i in range(self.v.shape[-1])], dim=-1
        )

    def heatflux(self, f=None):
        if f is None:
            f = self.f
        m0 = self._m0(f)
        m1 = self._m1(f)
        m2 = self._m2(f)
        m3 = self._m3(f)
        u = m1 / m0
        q = (-(u ** 3) * m0 + 3 * (u ** 2) * m1 - 3 * u * m2 + m3)
        return q

    def rho_u_theta(self, f=None):
        # TODO: 这里可以避免重复计算
        return self.density(f), self.velocity(f), self.temperature(f)

    def maxwellian(self, rho_u_theta=None):
        if rho_u_theta is None:
            rho_u_theta = self.rho_u_theta()
        rho, u, T = rho_u_theta
        return (rho / torch.sqrt(2 * math.pi * T) ** self.v.shape[-1]) * torch.exp(
            -((u[..., None, :] - self.v) ** 2).sum(dim=-1) / (2 * T)
        )

    def maxwellian_half(self, rho_u_theta=None, sign=0):
        maxwellian=self.maxwellian(rho_u_theta = rho_u_theta)
        if sign > 0:
            maxwellian[..., self.v[...,0] < 0] = 0
            maxwellian[..., self.v[...,0] == 0] = maxwellian[..., self.v[...,0] == 0]/2
        elif sign < 0:
            maxwellian[..., self.v[...,0] > 0] = 0
            maxwellian[..., self.v[...,0] == 0] = maxwellian[..., self.v[...,0] == 0]/2
        else:
            raise ValueError("sign should be 1 or -1")
        return maxwellian

    def one_side(self, f=None, sign=0):
        if f is None:
            f=self.f
        f = f.clone()
        if sign == 1:
            f[..., self.v[...,0] < 0] = 0
            f[..., self.v[...,0] == 0] = f[..., self.v[...,0] == 0]/2
        elif sign == -1:
            f[..., self.v[...,0] > 0] = 0
            f[..., self.v[...,0] == 0] = f[..., self.v[...,0] == 0]/2
        else:
            raise ValueError("sign should be 1 or -1")
        return f

    def reverse(self,f=None):
        if f is None:
            f=self.f
        return torch.flip(f, [-1,])

    def from_HermiteDis(self, dis: HermiteDis) -> "DVDis":
        rho, u, theta = dis.rho_u_theta
        f = dis.coef
        v = (self.v[...,0] - u) / torch.sqrt(theta)
        h = 0
        sqr_T=theta.sqrt()
        kernel = torch.exp(-(v ** 2) / 2) / torch.sqrt(2 * math.pi * theta)
        for i in range(f.shape[-1]):
            h = h + f[..., i : i + 1] * eval_hermitenorm(i, v) / sqr_T ** i
        # h = (eval_hermitenorm_stack(f.shape[-1],v) @ f.unsqueeze(-1)).squeeze(-1)
        self.f = h * kernel
        return self


def uniform_velocity(
    nv: int, vmin: float, vmax: float
) -> Tuple[torch.Tensor, torch.Tensor]:
    """return a discrete of velocity with uniform partition.

    v_0=vmin+0.5*(vmax-vmin)/nv
    v_{nv-1}=vmax-0.5*(vmax-vmin)/nv

    Args:
        nv (int): Numbers of grid
        vmin (float): vmin
        vmax (float): vmax

    Returns:
        Tuple[torch.Tensor,torch.Tensor]: v, weight of v
    """
    dv = 2 / nv
    v = torch.arange(start=-1 + 0.5 * dv, end=1 + 0.5 * dv, step=dv)
    w = dv * torch.ones_like(v)
    v1 = 0.5 * (v + 1.0) * (vmax - vmin) + vmin
    w1 = 0.5 * (vmax - vmin) * w
    #v1 = torch.tensor(v1, dtype=torch.get_default_dtype())
    #w1 = torch.tensor(w1, dtype=torch.get_default_dtype())
    v1 = v1.reshape([-1, 1])
    w1 = w1.reshape([-1, 1])
    return v1, w1


def legendre_velocity(
    nv: int, vmin: float, vmax: float
) -> Tuple[torch.Tensor, torch.Tensor]:
    """return a discrete of velocity with points on legendre-gauss points.

    Args:
        nv (int): Numbers of grid
        vmin (float): vmin
        vmax (float): vmax

    Returns:
        Tuple[torch.Tensor,torch.Tensor]: v, weight of v
    """
    v, w = np.polynomial.legendre.leggauss(nv)
    v1 = 0.5 * (v + 1.0) * (vmax - vmin) + vmin
    w1 = 0.5 * (vmax - vmin) * w
    v1 = torch.tensor(v1, dtype=torch.get_default_dtype())
    w1 = torch.tensor(w1, dtype=torch.get_default_dtype())
    v1 = v1.reshape([-1, 1])
    w1 = w1.reshape([-1, 1])
    return v1, w1


def uniform_velocity_2d(nv, vmin, vmax):
    v1, w1 = uniform_velocity(nv[0], vmin[0], vmax[0])
    v2, w2 = uniform_velocity(nv[1], vmin[1], vmax[1])
    grid_x, grid_y = torch.meshgrid(v1.squeeze(), v2.squeeze())
    v = torch.stack([grid_x, grid_y], dim=-1)
    w = w1.reshape([-1, 1]) * w2.reshape([1, -1])
    v = v.reshape([-1, 2])
    w = w.reshape([-1, 1])
    return v, w


def legendre_velocity_2d(nv, vmin, vmax):
    v1, w1 = legendre_velocity(nv[0], vmin[0], vmax[0])
    v2, w2 = legendre_velocity(nv[1], vmin[1], vmax[1])
    grid_x, grid_y = torch.meshgrid(v1.squeeze(), v2.squeeze())
    v = torch.stack([grid_x, grid_y], dim=-1)
    w = w1.reshape([-1, 1]) * w2.reshape([1, -1])
    v = v.reshape([-1, 2])
    w = w.reshape([-1, 1])
    return v, w

def ND_velocity(func:Callable[[int,int,int],Tuple]):
    def wrapper(nvT:Tuple, vminT:Tuple, vmaxT:Tuple):
        assert len(nvT)==len(vminT)
        assert len(nvT)==len(vmaxT)
        vL=[]
        wL=[]
        DIM=len(nvT)
        for d,nv,vmin,vmax in zip(range(DIM), nvT, vminT, vmaxT):
            v, w = func(nv, vmin, vmax)
            w=w.reshape([1,]*d+[-1,]+[1,]*(DIM-d-1))
            vL.append(v)
            wL.append(w)
        grids = torch.meshgrid(*[v.squeeze() for v in vL])
        V = torch.stack(grids, dim=-1)
        W = functools.reduce(lambda a,b:a*b,wL)
        V = V.reshape([-1, DIM])
        W = W.reshape([-1, 1])
        return V, W
    return wrapper

uniform_velocity_nd=ND_velocity(uniform_velocity)
legendre_velocity_nd=ND_velocity(legendre_velocity)

def velocity_list(func, nvT:Tuple, vminT:Tuple, vmaxT:Tuple):
    assert len(nvT)==len(vminT)
    assert len(nvT)==len(vmaxT)
    vL=[]
    wL=[]
    DIM=len(nvT)
    for d,nv,vmin,vmax in zip(range(DIM), nvT, vminT, vmaxT):
        v, w = func(nv, vmin, vmax)
        w=w.reshape([1,]*d+[-1,]+[1,]*(DIM-d-1))
        vL.append(v)
        wL.append(w) 
    return vL,wL

def product_meshgrid(vL:List, wL:List):
    assert len(vL)==len(wL)
    DIM = len(vL)
    grids = torch.meshgrid(*[v.squeeze() for v in vL])
    V = torch.stack(grids, dim=-1)
    W = functools.reduce(lambda a,b:a*b,wL)
    V = V.reshape([-1, DIM])
    W = W.reshape([-1, 1])
    return V, W

class DVDis_Chu(distributionBase):
    def __init__(self, v_meta, data_g, data_h, ndim=3):
        self.v_meta = v_meta
        self.ndim = ndim
        self.g = DVDis(v_meta, data_g)
        self.h = DVDis(v_meta, data_h)

    def __add__(self,another):
        if isinstance(another,DVDis_Chu):
            return DVDis_Chu(self.v_meta,self.g.f+another.g.f,self.h.f+another.h.f,ndim=self.ndim)
        else:
            return DVDis_Chu(self.v_meta,self.g.f+another,self.h.f+another,ndim=self.ndim)

    def __radd__(self,another):
        return DVDis_Chu(self.v_meta,self.g.f+another,self.h.f+another,ndim=self.ndim)

    def __sub__(self,another):
        return DVDis_Chu(self.v_meta,self.g.f-another.g.f,self.h.f-another.h.f,ndim=self.ndim)

    def __mul__(self,multiplier):
        return DVDis_Chu(self.v_meta,self.g.f*multiplier,self.h.f*multiplier,ndim=self.ndim)

    def __rmul__(self,multiplier):
        return DVDis_Chu(self.v_meta,self.g.f*multiplier,self.h.f*multiplier, ndim=self.ndim)

    def __truediv__(self,divider):
        return DVDis_Chu(self.v_meta,self.g.f/divider,self.h.f/divider, ndim=self.ndim)

    @property
    def v(self):
        return self.v_meta.v

    @property
    def v2(self):
        return self.v_meta.v2

    @property
    def v3(self):
        return self.v_meta.v3

    @property
    def w(self):
        return self.v_meta.w

    def density(self, g=None, h=None):
        if g is None:
            g = self.g.f
        if h is None:
            h = self.h.f
        return self.g.density(f=g)

    def velocity(self, g=None, h=None):
        if g is None:
            g = self.g.f
        if h is None:
            h = self.h.f
        return self.g.velocity(f=g)

    def temperature(self, g=None, h=None):
        if g is None:
            g = self.g.f
        if h is None:
            h = self.h.f
        temp = (
            self.g.temperature(f=g) + 2 * self.h.density(f=h) / self.g.density(f=g)
        ) / self.ndim
        return temp

    def rho_u_theta(self, g=None, h=None):
        rho = self.density()
        u = self.velocity()
        T = self.temperature()
        return rho, u, T

    def maxwellian(self, g=None, h=None, rho_u_theta=None):
        if rho_u_theta is None:
            if g is None:
                g = self.g.f
            if h is None:
                h = self.h.f
            rho, u, T = self.rho_u_theta(g=g, h=h)
        else:
            rho, u, T = rho_u_theta
        gm = self.g.maxwellian(rho_u_theta=(rho, u, T))
        hm = gm * T *(self.ndim-self.g.ndim)/2
        return DVDis_Chu(self.v_meta, gm, hm)


class HermiteDisND(HermitedistributionBase):
    """
    HermiteDis: Gas distribution in Hermite form.
    """

    def __init__(self, u:torch.Tensor, theta:torch.Tensor, coef:torch.Tensor, indt:index_table):
        self.u = u
        self.theta = theta
        self.coef = coef
        self.indt = indt
        self.D = indt.DIM
        self.DIM = self.D
        
    def new(self,u,theta,coef,indt=None):
        if indt is None:
            indt=self.indt
        return HermiteDisND(u,theta,coef,indt)
        
    def M(self):
        return self.indt.M

    def __str__(self) -> str:
        return "u:"+self.u.__str__()+"\n"+"theta:"+self.theta.__str__()+"\n"+"coef:"+self.coef.__str__()

    @staticmethod
    def empty(order: int,dim=3, device='cpu') -> "HermiteDisND":
        indt = index_table(order,dim)
        u = torch.empty((0, dim),device=device)
        theta = torch.empty((0, 1),device=device)
        coef = torch.empty((0, len(indt.iNto1)),device=device)
        return HermiteDisND(u, theta, coef, indt)

    def from_DVDis(self, dis: "DVDis", u=None,theta=None) -> "HermiteDis":
        pass

    def cpu(self):
        return HermiteDisND(self.u.cpu(), self.theta.cpu(), self.coef.cpu(), self.indt)

    def cuda(self):
        return HermiteDisND(self.u.cuda(), self.theta.cuda(), self.coef.cuda(), self.indt)

    def to(self, args):
        return HermiteDisND(self.u.to(args), self.theta.to(args), self.coef.to(args), self.indt)

    @property
    def ORDER(self):
        return self.indt.ORDER

    @property
    def rho(self):
        return self.coef[..., 0:1]

    def to_tensor(self):
        return torch.cat([self.u, self.theta, self.coef], dim=-1)

    def density(self):
        return self.coef[...,0:1]

    def velocity(self):
        return self.u

    def temperature(self):
        return self.theta
    
    @property
    def moments(self):
        # TODO: 返回0、1、2阶矩，其中0阶矩有1个，1阶矩有3个，2阶矩有9个
        u, theta, coef = self.u, self.theta, self.coef
        v0 = coef[..., 0:1]
        f1=coef[..., 1:self.DIM+1]
        v1 = u * v0 + f1
        u2 = u * u
        v2 = (self.DIM * theta + u2.sum(dim=-1,keepdim=True))*v0+2*(u*f1).sum(dim=-1,keepdim=True)+2*(
           coef[...,self.indt.v2].sum(dim=-1,keepdim=True))
        v2 = v2 / 2
        return v0, v1, v2

    @property
    def rho_u_theta(self):
        m0, m1, m2 = self.moments
        v0 = m0
        v1 = m1 / m0
        v2 = (2.0 * m2 / v0 - (v1 * v1).sum(dim=-1,keepdim=True))/self.D
        return v0, v1, v2

    def as_tuple(self):
        return self.u, self.theta, self.coef

    def __add__(self, anotherDis):
        return HermiteDisND(self.u, self.theta, self.coef + anotherDis.coef, self.indt)

    def __radd__(self, anotherDis):
        return HermiteDisND(self.u, self.theta, self.coef + anotherDis.coef, self.indt)

    def __sub__(self, anotherDis):
        return HermiteDisND(self.u, self.theta, self.coef - anotherDis.coef, self.indt)

    def __mul__(self, multiplier):
        return HermiteDisND(self.u, self.theta, multiplier * self.coef, self.indt)

    def __rmul__(self, multiplier):
        return HermiteDisND(self.u, self.theta, multiplier * self.coef, self.indt)

    def __truediv__(self, divider):
        return HermiteDisND(self.u, self.theta, self.coef / divider, self.indt)

    def roll(self, idx):
        return HermiteDisND(
            d3roll(self.u, idx), d3roll(self.theta, idx), d3roll(self.coef, idx), self.intd
        )

    def clone(self):
        return HermiteDisND(self.u.clone(), self.theta.clone(), self.coef.clone(), self.indt)

    def down(self):
        indt=index_tables.get(self.indt.ORDER-1,self.indt.DIM)
        return HermiteDisND(self.u, self.theta, self.coef[..., :indt.len], indt)

    def Project_to_STD(self, STEP: int = 1):
        """
        Project_to_STD 将分布参数重设，重新投影到最佳参数上
        """
        _, u, theta = self.rho_u_theta
        dis = self.Project_RK4(u, theta, STEP=STEP)
        return dis

    def Project_2B(self, dest: "HermiteDisND", STEP: int = 1) -> "HermiteDis":
        """
        Project_A2B 将分布投影到dest对应的参数上
        """
        u, theta = dest.u, dest.theta
        dis = self.Project_RK4(u, theta, STEP=STEP)
        return dis

    def mulvec(self, axis = 0):
        """axis=0 for x, 1 for y, 2 for z.
        """
        tmp = torch.zeros_like(self.coef)
        
        coef = self.coef
        ind= self.indt.m1[axis]
        #ind = torch.tensor(self.indt.get_order(self.indt.u1[axis][0].tolist(),dim=axis),device=tmp.device) + 1 # 1,2,....,
        ind = ind.reshape([1, 1, -1]).to(device=coef.device)
        #print(ind.shape)
        
        tmp[..., :] = tmp[..., :] + coef[..., :] * self.u[...,axis:axis+1]
        tmp[..., self.indt.d1[axis][0]] = tmp[..., self.indt.d1[axis][0]] + coef[..., self.indt.d1[axis][1]] * self.theta
        tmp[..., self.indt.u1[axis][0]] = tmp[..., self.indt.u1[axis][0]] + coef[..., self.indt.u1[axis][1]] * ind
        
        return HermiteDisND(self.u, self.theta, tmp,self.indt)
    

    def Project_RK4_helper(self, dis, u0, theta0, u, theta):
        tmp=[torch.zeros_like(dis) for d in range(self.DIM)]
        tmpt=[torch.zeros_like(dis) for d in range(self.DIM)]
        for d in range(self.DIM):
            #print(self.indt.d2[d][0],self.indt.d2[d][1])
            #print(tmp[0].shape)
            tmp[d][...,self.indt.d1[d][0]]=dis[...,self.indt.d1[d][1]]
            tmpt[d][...,self.indt.d2[d][0]]=dis[...,self.indt.d2[d][1]]
        ud = (u0 - u)
        #print(ud.shape,tmp[0].shape,theta0.shape,theta.shape,tmpt[0].shape)
        tmp = sum([ud[...,d:d+1]*tmp[d] for d in range(self.DIM)]) + 0.5 * (theta0 - theta) * sum(tmpt)
        return tmp

    def Project_RK4_step(self, dis, u0, theta0, u, theta, h: float):
        k1 = self.Project_RK4_helper(dis, u0, theta0, u, theta)
        k2 = self.Project_RK4_helper(dis + h / 2 * k1, u0, theta0, u, theta)
        k3 = self.Project_RK4_helper(dis + h / 2 * k2, u0, theta0, u, theta)
        k4 = self.Project_RK4_helper(dis + h * k3, u0, theta0, u, theta)
        dis = dis + (k1 + 2.0 * k2 + 2.0 * k3 + k4) * h / 6.0
        return dis

    def Project_RK4(self, u, theta, STEP: int = 1):
        """Project_RK4 使用R-K 4格式对分布函数进行投影

        Parameters
        ----------
        dis : torch.Tensor
            分布函数
        u : torch.Tensor
            分布中心
        theta : torch.Tensor
            分布宽度
        STEP : int, optional
            RK方法的步数, by default 1
        Returns
        -------
        torch.Tensor
            投影后的分布函数
        """
        h = 1.0 / STEP
        tmpslice = self.coef
        u0 = self.u
        theta0 = self.theta
        for i in range(STEP):
            tmpslice = self.Project_RK4_step(tmpslice, u0, theta0, u, theta, h)
        return HermiteDisND(u, theta, tmpslice, self.indt)


import torch

def eval_hermitenorm(n: int, x: torch.Tensor):
    if n < 0:
        return float("nan") * torch.ones_like(x)
    elif n == 0:
        return torch.ones_like(x)
    elif n == 1:
        return x
    else:
        y3 = torch.zeros_like(x)
        y2 = torch.ones_like(x)
        for k in range(n, 1, -1):
            y1 = x * y2 - k * y3
            y3 = y2
            y2 = y1
        return x * y2 - y3

def eval_hermitenorm_seq(n: int, x: torch.Tensor):
    y=torch.zeros((n+1,)+x.shape)
    if n < 0:
        raise ValueError
    if n >= 0:
        y[0]=torch.ones_like(x)
    if n >= 1:
        y[1]=x
    if n >= 2:
        for j in range(2,n+1):
            y[j]=x*y[j-1]-(j-1)*y[j-2]
    return y


import torch
from math import factorial
from .operators import make_tensor
from bte.grad.distribution import HermiteDisND
from .specials import eval_hermitenorm_seq

class DisTranfer():
    def __init__(self,v_meta,indt):
        assert len(v_meta.v_dims)==indt.DIM
        self.v_meta=v_meta
        self.indt=indt
        self.hermite={}
        self.ORDER=indt.ORDER
        self.DIM=indt.DIM
        self.v_shape=tuple([len(v.v) for v in v_meta.v_dims])
        for d,vm in enumerate(self.v_meta.v_dims):
            self.hermite[d]=eval_hermitenorm_seq(self.ORDER,vm.v.cpu()).permute(1,2,0).squeeze()
        self.M=torch.zeros((len(self.indt.i1toN),)+self.v_shape)
        for i,orders in self.indt.i1toN.items():
            #在这里构造相应的Hermite多项式的值
            self.M[i]=self.getHermite(orders)
        self.M=self.M.to(device=v_meta.v_dims[0].v.device)
        
    def getHermite(self,order):
        ans=torch.zeros(self.v_shape)
        fs=[self.hermite[d][...,di]/factorial(di)*self.v_meta.v_dims[d].w.cpu().flatten() for d,di in enumerate(order)]
        ans=make_tensor(fs)
        return ans
            
    def DVMtoHermite(self,DVDis):
        # 暂时只支持DIM=1,2,3
        f=DVDis.f
        f=f.reshape(f.shape[:-1]+(self.v_shape))
        if self.DIM==3:
            coef=torch.einsum("...ijk,cijk->...c",f,self.M)
        elif self.DIM==2:
            coef=torch.einsum("...ij,cij->...c",f,self.M)
        elif self.DIM==1:
            coef=torch.einsum("...i,ci->...c",f,self.M)
        else:
            raise ValueError
        gdis=HermiteDisND(torch.zeros(f.shape[:-self.DIM]+(self.DIM,),device=f.device),torch.ones(f.shape[:-self.DIM]+(1,),device=f.device),coef,self.indt)
        return gdis

    def HermitetoDVM(self,gdis):
        pass


import torch
import numpy as np

DATAPATH="/home/zyli/NRXXPKU/examples/test33/A/A_10_10.bin"
MAX_DEG=10

def load_collision_kernel(kernel_path=DATAPATH,  max_order=MAX_DEG, order=None,dtype=torch.float32):
    assert (order is None) or isinstance(order,int)
    if isinstance(order,int):
        assert order<=max_order

    n_mom=(max_order + 1) * (max_order + 2) * (max_order + 3) // 6
    nn=(order + 1) * (order + 2) * (order + 3) // 6
    with open(kernel_path) as f:
        bdata = np.fromfile(f, dtype=np.float64)
    s= bdata [5 * n_mom * n_mom + 5 * n_mom];
    s = s+ bdata[5 * n_mom * n_mom + 5];
    s = -1/s
    bdata=bdata*s
    bdata=bdata.reshape([n_mom,n_mom,n_mom])

    if order is None:
        re=torch.from_numpy(bdata)
    else:
        re=torch.from_numpy(bdata[:nn,:nn,:nn])
    return re.to(dtype=dtype)

def collision_f(Kernel,f):
    fs=f.shape
    f=f.reshape([-1,f.shape[-1]])
    q=torch.einsum("kij,Ni,Nj->Nk",Kernel,f,f)
    q=q.reshape(fs)
    return q